#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import random
import string
import math
import numpy as np
import timeit
from collections import Counter

from collections import defaultdict

from pyyacp.profiler import ColumnProfiler, ColumnByCellProfiler
from pyyacp.profiler import is_not_empty
from outliers import smirnov_grubbs as grubbs

import structlog
log =structlog.get_logger()


def get_outlier_indices(data):
    num_data = []
    for d in data:
        try:
            num_data.append(float(d))
        except Exception as e:
            pass
    if len(num_data) < len(data):
        log.warn('Could only convert ' + str(len(num_data)) + ' of ' + str(len(data)) + ' values in column')
    return grubbs.min_test_indices(num_data, alpha=0.05)

class OutlierProfiler(ColumnProfiler,ColumnByCellProfiler):

    def __init__(self):
        super(OutlierProfiler, self).__init__('cdp', 'c_dist')
        self.dv = defaultdict(int)

    def profile_column(self, column, meta):
        return self._profile_column(column, meta)

    def result(self):
        return self._compile_stats()

    def accept(self, cell):
        if is_not_empty(cell):
            for c in cell:
                self.dv[c] += 1

        #self.values.append(cell)

    def _profile_column(self, values, meta):
        data_type = meta['data_type']
        if data_type=='int':
            num_values = map(int, values)
        elif data_type=='float':
            num_values = map(float, values)

    def _compile_stats(self):
        return {'dist':dict(self.dv)}



# -*- coding: utf-8 -*-

"""Console script for pyyacp."""
import sys
import click

from pyyacp import datatable
from pyyacp.profiler.profiling import apply_profilers
from pyyacp.table_structure_helper import AdvanceStructureDetector

from pyyacp import YACParser

import click

from pyyacp.web.to_html import to_html_string


class NotRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop('not_required_if')
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs['help'] = (kwargs.get('help', '') +
            ' NOTE: This argument is mutually exclusive with %s' %
            self.not_required_if
        ).strip()
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.not_required_if in opts

        if other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`" % (
                        self.name, self.not_required_if))
            else:
                self.prompt = None

        return super(NotRequiredIf, self).handle_parse_result(
            ctx, opts, args)

@click.group()
def cli():
    """Console script for pyyacp."""
    pass

@cli.command()
@click.option('-f','--file',  help='File', cls=NotRequiredIf, not_required_if='url')
@click.option('-u','--url',  help='URL'  , cls=NotRequiredIf, not_required_if='file')
def inspect(file, url):
    """Inspect a CSV file to figure out about the dialect, comment and header lines and the overall structure."""

    structure_detector = AdvanceStructureDetector()
    sample_size = 1800

    click.echo("Input filename:{}, ur:{}".format(file, url))
    yacp = YACParser(filename=file, url=url, structure_detector=structure_detector, sample_size=sample_size)
    if url is None:
        url = 'http://example.org/table'
    tables = datatable.parseDataTables(yacp, url=url, max_tables=10)

    click.echo("Found {} tables".format(len(tables)))
    for table in tables:
        table.print_summary()
    return 0

@cli.command()
@click.option('-f','--file',  help='File', cls=NotRequiredIf, not_required_if='url')
@click.option('-u','--url',  help='URL'  , cls=NotRequiredIf, not_required_if='file')
def clean(file, url):
    """Parse and clean a CSV file (strip comments, utf-8 encoding, default dialect"""


    structure_detector = AdvanceStructureDetector()
    sample_size = 1800

    click.echo("Input filename:{}, ur:{}".format(file, url))
    yacp = YACParser(filename=file, url=url, structure_detector=structure_detector, sample_size=sample_size)

    if url is None:
        url = 'http://example.org/table'
    tables = datatable.parseDataTables(yacp, url=url, max_tables=10)

    if len(tables)>1:
        click.echo("Found {} tables".format(len(tables)))
    else:
        print(tables[0].generate())

@cli.command()
@click.option('-f','--file',  help='File', cls=NotRequiredIf, not_required_if='url')
@click.option('-u','--url',  help='URL'  , cls=NotRequiredIf, not_required_if='file')
@click.option('--html',  help='html representation',type=click.Path(resolve_path=True))
@click.option('-l','--load',  help='try to open/load html file',count=True)
@click.option('-s','--sample',  help='sample rows',type=int, default=None)

def profile(file, url, html, load, sample):
    """Console script for pyyacp."""

    structure_detector = AdvanceStructureDetector()
    sample_size = 1800
    max_tables=1


    click.echo("Input filename:{}, ur:{}".format(file, url))
    yacp = YACParser(filename=file, url=url, structure_detector=structure_detector, sample_size=sample_size)
    if url is None:
        url='http://example.org/table'
    tables = datatable.parseDataTables(yacp, url=url, max_tables=max_tables)

    if max_tables==1:
        table=tables
    elif len(tables)>1:
        click.echo("Found {} tables".format(len(tables)))
        return


    apply_profilers(table)


    if html:
        click.echo("Writing HTML representation to {}".format(click.format_filename(html)))
        with open(html, "w") as f:
            f.write(to_html_string(table,sample=sample))
        if load:
            click.launch(html)
    else:
        table.print_summary()

if __name__ == "__main__":
    cli() # pragma: no cover



import functools
from time import sleep

from faker import Factory
fake = Factory.create()
from pyyacp.column_format_detector import pattern_comparator, translate




def filter1(value):
    return len(value)!=0

## digits
def lower(value):
    return value.lower()


def count(value):
    sleep(0.001)
    return value


def random_number(size, digits=None, fix_len=False):
    return [ unicode(fake.random_number(digits=digits, fix_len=fix_len)) for i in range(0,size)]


int_columns=[
    random_number(10,digits=1, fix_len=True),
    random_number(10,digits=2, fix_len=True),
    random_number(10,digits=2)
]

inputs=[unicode(fake.name()) for i in range(0,1000)]

print inputs
def translate1(inputs):

    inputs = [i for i in inputs if filter1(i)]
    inputs = [lower(i) for i in inputs]
    inputs = [count(i) for i in inputs]

    p = [translate(i) for i in inputs if len(i.strip()) > 0]
    p = sorted(p, key=functools.cmp_to_key(pattern_comparator))
    return p

def translate2(inputs):
    p=[]
    for i in inputs:
        if filter1(i):
            i=lower(i)
            i=count(i)
            p.append(translate(i))
    p = sorted(p, key=functools.cmp_to_key(pattern_comparator))

    return p

def translate4(inputs):
    p=[]
    pa=p.append
    for i in inputs:
        if filter1(i):
            i=lower(i)
            i=count(i)
            pa(translate(i))
    p = sorted(p, key=functools.cmp_to_key(pattern_comparator))

    return p

def translate3(inputs):
    p=map(lower,filter(lambda x: filter1(x), inputs))
    p = map(count, p)
    p=map(translate, p)
    p = sorted(p, key=functools.cmp_to_key(pattern_comparator))

    return p

if __name__ == '__main__':
    import timeit

    repeat=1
    numer=1
    print "list_comp :", timeit.Timer('f(inputs)', 'from __main__ import inputs, translate1 as f').repeat(repeat=repeat, number=numer)
    print "loop :", timeit.Timer('f(inputs)', 'from __main__ import inputs, translate2 as f').repeat(repeat=repeat, number=numer)
    print "loop2 :", timeit.Timer('f(inputs)', 'from __main__ import inputs, translate4 as f').repeat(repeat=repeat, number=numer)
    print "map_filter :", timeit.Timer('f(inputs)', 'from __main__ import inputs, translate3 as f').repeat(repeat=repeat, number=numer)




from pyyacp import YACParser
from pyyacp import datatable
from pyyacp.cli import inspect
from pyyacp.table_structure_helper import AdvanceStructureDetector

filename="/Users/jumbrich/new_folder/dev/tablerec/tablerec/library_data.csv"
filename=None
url='http://data.wu.ac.at/data/unibib/library_data.csv'



#import requests
#r = requests.get(url)
#iter_lines = [line for line in r.iter_lines(chunk_size=1024)]
#split_lines = r.content.splitlines()


def my_iter_lines(result):
    buf = b''
    for chunk in result.iter_content(chunk_size=64 * 1024, decode_unicode=False):
        buf += chunk
        pos = 0
        while True:
            eol = buf.find(b'\n', pos)
            if eol != -1:
                yield buf[pos:eol]
                pos = eol + 1
            else:
                buf = buf[pos:]
                break
    if buf:
        yield buf

#iter1_lines=[line for line in my_iter_lines(r)]
#print(len(iter_lines), len(split_lines), len(iter1_lines))
#import sys
#sys.exit(0)

filename='/Users/jumbrich/new_folder/data/catalog/data/data.wu.ac.at/75/L2RhdGEvdW5pYmliL2xpYnJhcnlfZGF0YS5jc3Y=/b48f86a2552739bf4e0503ff7ba64cd1.gz'


url=None
import os
#print(os.stat(filename))
#print(os.path.isfile(filename))
print(open(filename))
structure_detector = AdvanceStructureDetector()
sample_size = 1800

yacp = YACParser(filename=filename, url=url, structure_detector=structure_detector, sample_size=sample_size)
if url is None:
    url = 'http://example.org/table'
tables = datatable.parseDataTables(yacp, url=url, max_tables=10)

for table in tables:
    table.print_summary()



# -*- coding: utf-8 -*-
import codecs
import unittest

from pyyacp.table_structure_helper import AdvanceStructureDetector, SimpleStructureDetector
from pyyacp.pyyacp import YACParser


class TestDescriptionDetectionAdvanceStructureDetector(unittest.TestCase):

    #structure_detector = AdvanceStructureDetector()

    def setUp(self):
        self.structure_detector = AdvanceStructureDetector()




    def test_no_comment(self):
        table=u"""h1,h2,h3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(0,len(yacp.descriptionLines))


    def test_empty_comment(self):
        table=u"""
                  h1,h2,h3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(1,len(yacp.descriptionLines))


    def test_one_comment(self):
        table=u"""t
                  h1,h2,h3
                  c1,c2,c3
                  c21,c22,c23
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(1,len(yacp.descriptionLines))

    def test_one_comment1(self):
        table=u"""t,
                  h1,h2,h3
                  c1,c2,c3
                  c21,c22,c23
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(1,len(yacp.descriptionLines))

    def test_one_comment1(self):
        table=u"""t,,
                  h1,h2,h3
                  c1,c2,c3
                  c21,c22,c23
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(1,len(yacp.descriptionLines))

    def test_mix_comment1(self):
        table=u"""
                  t,,

                  h1,h2,h3
                  c1,c2,c3
                  c21,c22,c23
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3
                  c1,c2,c3"""
        yacp = YACParser(content=table, sample_size=1800,structure_detector=self.structure_detector)
        self.assertEquals(3,len(yacp.descriptionLines))

class TestDescriptionDetectionSimpleStructureDetector(TestDescriptionDetectionAdvanceStructureDetector):
    def setUp(self):
        self.structure_detector = SimpleStructureDetector()

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import codecs
import unittest

from pyyacp.pyyacp import YACParser


def generate_csv(encoding):
    print "e",encoding
    data=u'ä,ö,ê\na,b,c'

    if encoding!='utf8':
        data=data.encode(encoding, errors='replace')
    print data
    return data


class TestEncodingDetection(unittest.TestCase):

    def test_utf8(self):
        encoding='utf8'
        yacp = YACParser(content=generate_csv(encoding), sample_size=1800)
        self.assertEquals(encoding,yacp.meta['encoding'])

    def test_ascii(self):
        encoding='us-ascii'
        yacp = YACParser(content=generate_csv(encoding), sample_size=1800)
        self.assertEquals(encoding, yacp.meta['encoding'])

    def test_latin(self):
        encoding='iso-8859-1'
        yacp = YACParser(content=generate_csv(encoding), sample_size=1800)
        self.assertEquals(encoding, yacp.meta['encoding'])

if __name__ == '__main__':
    unittest.main()




from jinja2 import Environment, FileSystemLoader, select_autoescape
import os

PATH = os.path.dirname(os.path.abspath(__file__))
env = Environment(
    autoescape=False,
    loader=FileSystemLoader(os.path.join(PATH, 'templates'))
)



def to_html(table, cnt, dir):
    m_order=['data_type','data_class','patterns',
             'stats_num_rows','stats_distinct','stats_empty',
             'stats_min_len','stats_mean_len','stats_max_len',
             'stats_constancy','stats_uniqueness','stats_top_value',
             'c_dist_dist','benford_is','benford_chi','benford_dist']
    output = os.path.join('.', "../html/{}.html".format(cnt))
    with open(output, "w") as f:
        template = env.get_template('profile.jinja')
        f.write(template.render(
            basic=table.table_structure(),
            columnIDs=table.columnIDs,
            tmeta=table.meta,
            m_order=m_order,
            comments=table.comments,
            data=table.data.head(5).to_html(classes='ui celled table'),
            meta=table.colum_profiles().to_dict(orient='index')#table.describe_colmeta().to_html(classes='ui celled table')
        ))

col_profile_keys = ['uri', 'digest', 'col',
          'stats_mean_len', 'stats_max_len', 'stats_distinct', 'stats_constancy', 'stats_min_len', 'stats_max_value',
          'stats_min_value',
          'data_type', 'patterns', 'stats_empty', 'stats_num_rows', 'benford_dist', 'stats_uniqueness', 'c_dist_dist',
          'benford_is', 'benford_chi', 'stats_top_value', 'data_class']


_table_structure=['uri','rows','columns','headers','comments','comment_lines','header_lines']
_table_profile = ['encoding', 'uri', 'filename', 'delimiter', 'quotechar', 'lineterminator', 'skipinitialspace',
                          'quoting', 'delimiter', 'doublequote', 'fd']

def to_html_string(table, sample=5):
    # table_meta={}
    col_profiles={}
    # for k, v in table.table_profile().items():
    #     table_meta[k] = v
    # for k, v in table.table_structure().items():
    #     table_meta[k] = v
    #
    for col, meta in table.colum_profiles().to_dict().items():
         col_profiles[col] = {k: meta[k] for k in col_profile_keys if k in meta}
    #
    t_c_p = {}
    for col, profile in col_profiles.items():
         for k, v in profile.items():
             if k != 'col':
                 d = t_c_p.setdefault(k, {})
                 d[col] = v
    # t_s = {k: v for k, v in table_meta.items() if k in _table_structure}
    # t_p = {k: v for k, v in table_meta.items() if k in _table_profile}
    #
    # t1 = table
    # t1_s = t_s
    # t1_p = t_p
    # t1_c_p = t_c_p
    #
    #
    if sample:
        data=    [list(table.data.columns)] + [list(row) for idx, row in table.data.head(sample).iterrows()]
    else:
        data = [list(table.data.columns)] + [list(row) for idx, row in table.data.iterrows()]

    html_data = {
        'basic':table.table_structure(),
        'columnIDs':table.columnIDs,
        'tmeta':table.meta,
        'comments':table.comments,
        'data':data,
        'meta':t_c_p
    }

    template = env.get_template('profile.jinja')
    return template.render(**html_data)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pyyacp.profiler import ColumnProfiler
from pyyacp.profiler.data_type_detection import DATETIME, INT, UNICODE

import re
SENT=re.compile('^(C[c])+( [Cc]+)*[\.!?]?$')
ENT=re.compile('^(C[c])+( [Cc]+){2,3}$')

class DataTypeInterpretation(ColumnProfiler):

    def __init__(self):
        super(DataTypeInterpretation, self).__init__('dclass','data_class')
        self.values=[]

    def profile_column(self, column, meta):
        data_class = 'UNDEF'
        if 'patterns' in meta and 'stats_max_len' in meta:
            return  self._analyse_ColumnPattern(column,meta)

        return data_class

    def _analyse_ColumnPattern(self, column,meta):
        patterns= meta['patterns']
        data_type=meta['data_type']
        #print meta
        min, mean, max = meta['stats_max_len'], meta['stats_mean_len'], meta['stats_min_len']
        fixed_len = min ==max
        unique = meta['stats_distinct'] + meta['stats_empty'] == meta['stats_num_rows']

        if data_type==DATETIME:
            return data_type.upper()

        if meta['stats_empty'] == meta['stats_num_rows']:
            return "EMPTY"

        s=''
        if data_type == INT:
            s+='NUM'
        elif data_type == UNICODE:
            if len(patterns)==1:
                if SENT.match(patterns[0][0]):
                    s+='SENTENCE'
                elif ENT.match(patterns[0][0]):
                    s += 'ENTITY'
                else:
                    s += 'UNI'
            else:
                s += 'UNI'
        else:
            s = 'UNDEF'

        if fixed_len:
            s+='_ID'
        else:
            s+='_VAR'

        if unique:
            s+='_UNIQ'

        if meta['stats_distinct'] == 1:
            s+="_SING"
        elif (meta['stats_num_rows'] > 10 and (meta['stats_distinct']/float(meta['stats_num_rows']))<0.1) \
                or (meta['stats_num_rows'] < 10 and meta['stats_distinct']<=2):
            s+='_CAT'
        return s



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import operator
from collections import defaultdict

import numpy as np

from pyyacp.profiler import ColumnProfiler, ColumnByCellProfiler
from pyyacp.timer import Timer, timer

__author__ = 'nina'

class ColumnStatsProfiler(ColumnProfiler,ColumnByCellProfiler):

    def __init__(self):
        super(ColumnStatsProfiler, self).__init__('csp', 'stats')
        self.vlen = []
        self.dv = defaultdict(int)

    @timer(key='csp_column')
    def profile_column(self, column, meta):
        return self._profile_column(column, meta)

    @timer(key='csp_result')
    def result(self):
        return self._compile_stats()

    def accept(self, cell):
        self.vlen.append(len(cell.strip()))
        self.dv[cell.strip()] += 1


    def _profile_column(self, values, meta):
        for v in values:
            self.vlen.append(len(v.strip()))
            self.dv[v.strip()]+=1
        return self._compile_stats()

    def _compile_stats(self):
        stats = {'num_rows': len(self.vlen)}

        a=np.array(self.vlen)
        an=a[a>0]

        if len(a[a==0])>0:
            self.dv.pop('')
        stats['min_value'] = min(self.dv) if len(self.dv)>0 else ''
        stats['max_value'] = max(self.dv) if len(self.dv)>0 else ''

        stats['min_len'] = min(an) if len(an)>0 else 0
        stats['max_len'] = max(an) if len(an)>0 else 0
        stats['mean_len'] = np.mean(an)
        stats['empty']= len(a[a==0])
        stats['distinct']=len(self.dv)
        stats['uniqueness']=len(self.dv)/float(len(a))

        sorted_values = sorted(self.dv.items(), key=operator.itemgetter(1), reverse=True)
        top_values = [(sorted_values[i][0], int(sorted_values[i][1])) for i in range(min(5, len(self.dv)))]


        stats['constancy']=max(self.dv.values())/float(len(a)) if len(self.dv)>0 and  len(a)>0 else 0
        stats['top_value'] = top_values[0] if len(top_values)>0 else None
        return stats


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import itertools

from pyyacp.profiler import simple_type_detection
from dtpattern import dtpattern1
DESCRIPTION_CONFIDENCE = 10

import structlog
log = structlog.get_logger()

class SimpleStructureDetector(object):
    # i would include emtpy lines for now
    def guess_description_lines(self,sample):
        return guess_description_lines(sample)
    # allow multple header lines if existing
    def guess_headers(self,sample, verbose=False):
        descriptionLines = self.guess_description_lines(sample)
        description = len(descriptionLines)
        if description:
            sample = sample[description:]
        emptyColumns = detect_empty_columns(sample)
        return guess_headers(sample, emptyColumns)

    def guess_columns(self,sample):
        descriptionLines = self.guess_description_lines(sample)
        description = len(descriptionLines)
        return len(sample[description])


ascii=set(['a','A','c','C'])
digits=set(['1','0'])
class AdvanceStructureDetector(object):

    # i would include emtpy lines for now
    def guess_description_lines(self, sample):
        r_len = [len(row) for row in sample]
        grouped_L = [(k, sum(1 for i in g)) for k, g in itertools.groupby(r_len)]
        est_colNo = _most_common_oneliner(r_len)
        c_lines=0
        cur_line = 0

        for i, group in enumerate(grouped_L):
            if group[0] < est_colNo:
                # We consider this as a comment line
                c_lines += group[1]
            else:
                # same length as the data rows, we should check if there are empty columns
                #if there are more than 50% empty columns, consider this a header line
                for lineno in range(cur_line, group[1] + cur_line):
                    empty_cells = 0.0
                    for a, cell in enumerate(reversed(sample[lineno])):
                        if len(cell.strip()) == 0 and empty_cells == a:
                            empty_cells += 1
                    r= empty_cells / len(sample[lineno]) if len(sample[lineno])>0 else 0
                    if  r > 0.5:
                        c_lines += 1
                    else:
                        break

            cur_line += group[1]
        return sample[0:c_lines]


    def _guess_header_by_group(self, groups, verbose=False):
        """

        :param groups: groups of csets rows
        :param verbose:
        :return:

        The idea is that the header line contains different characters as the remaining rows
        We translate each value to its cset print ( e.g. Berlin -> Cc, AT 10 -> C d)
        and compare the cset prints which each other

        Heuristcs are

        ## Simple cases first
        * If the prev row contaisn ascii and the next digits -> assume prev row is header
        *

        """
        if verbose:
            print(" _guess_header_by_group({})".format(groups))
        if len(groups)==1:
            ##asume no header
            return -1
        elif len(groups)==2:
            #two groups, very likely a header
            return groups[0][1]
        else:
            header_rows = -1

            prev_pat=None
            for i, pat in enumerate(groups):

                if prev_pat is None:
                    prev_pat=pat

                else:
                    if verbose:
                        print("  cell_pattern-{}: {}".format(i, pat))

                    ## compare this pattern with the previous one
                    ## we know that they have different csets

                    p_sym = set(pat[0])  # get unique cset symbols for this pattern
                    h_sym = set(prev_pat[0])  # current header symbols

                    new_sym = p_sym - h_sym  # > new set of cset symbols

                    _new_digits=bool(new_sym & digits) # are there digits in the new symbols
                    _has_ascii = bool(h_sym & ascii) # does the potential header have ascii characters?


                    # Do we have a header?
                    # 1) we have digits in the next rows
                    # pot header needs to have ascii and we add digits
                    digits_added = _has_ascii and len(new_sym) > 0 and _new_digits

                    # 2) new symbols
                    new_symbols = len(new_sym) > 0 and len(new_sym - h_sym)>0
                    if verbose:
                        print("   ascii in h:{},digits in new:{}".format(h_sym.issubset(ascii),
                                                                         len(new_sym) > 0 and _new_digits))
                        print("   new symbols added:{} subset:{}".format(new_symbols,new_sym.issubset(h_sym)))

                    is_header = (digits_added or new_symbols)
                    if verbose:
                        print("   header?: {} h:{} vs p{}, new:{}".format(is_header, h_sym, p_sym, new_sym))
                    if is_header:
                        if verbose:
                            print("   >>>>we think we have a header: {}".format(prev_pat))
                        header_rows = prev_pat[1]
                if header_rows != -1:
                    break

            return header_rows

    # allow multple header lines if existing
    def guess_headers(self, sample, verbose=False):
        descriptionLines = self.guess_description_lines(sample)
        #skip description lines
        sample = sample[len(descriptionLines):]
        #check how many row/col groups we still have
        r_len = [len(row) for row in sample]
        grouped_L = [(k, sum(1 for i in g)) for k, g in itertools.groupby(r_len)]
        est_colNo = _most_common_oneliner(r_len)

        if len(grouped_L)>0 and grouped_L[0][0] == est_colNo:
            ##lets assume that the first group is the one we use for the header detection
            ##also make sure that the length is the estimated col length

            ##convert the rows into columns
            cols = map(list, zip(*sample))
            guessed_col_header = []
            for col in cols:
                if verbose:
                    print ("Column:{}".format(col))
                pattern = dtpattern1.translate_all(col,sort=False)
                L1 = dtpattern1.l1_aggregate(pattern)
                symbols = dtpattern1.l2_aggregate(grouped=L1)
                d=[ (k[0],sum([cc[1] for cc in k[2]])) for k in symbols]
                p_guess_headers = self._guess_header_by_group(d,verbose=verbose)
                if verbose:
                    print (p_guess_headers)
                    print ('symbols:{}'.format(symbols))
                guessed_col_header.append(p_guess_headers)

            def mode(array):
                most = max(list(map(array.count, array)))
                return list(set(filter(lambda x: array.count(x) == most, array)))

            #TODO
            #This needs a log of thought and debugging
            mc = _most_common_oneliner([x for x in guessed_col_header if x>=0])
            if guessed_col_header.count(mc)==1:
                #hmm the most common appears only once,
                # lets check the minimum,
                if min(guessed_col_header)==1:
                    mc=1

            max_h = max(guessed_col_header) if len(guessed_col_header)>0 else 0
            #print sample
            if verbose:
                print ("most_common:{}, max:{}, modes:{}, {}".format(mc, max_h, mode(guessed_col_header), guessed_col_header))
            if mc >= 0:
                return sample[0: mc]
            return []

        else:
            if len(grouped_L)>0:
                raise ValueError("Header detectiong failed, no potential header group ( likely cause, too many empty cells -> description line)")
            else:
                raise ValueError(
                    "Header detectiong failed, no lines available")



    def guess_columns(self, sample):
        r_len = [len(row) for row in sample]
        return _most_common_oneliner(r_len)

def _most_common_oneliner(L):
    return max(itertools.groupby(sorted(L)),
               key=lambda x_v: (len(list(x_v[1])), -L.index(x_v[0]))
               )[0] if len(L) > 0 else 0


def guess_description_lines(sample):
    first_lines = True
    conf = 0
    description_lines = 0
    for i, row in enumerate(sample):
        if len(row) == 1 and len(row[0]) > 0:
            if not first_lines:
                # a description line candidate not in the first line, possibly no description line
                description_lines = 0
                break
        elif len(row) > 1 and all([len(c) == 0 for c in row[1:]]):
            if not first_lines:
                # a description line candidate not in the first line, possibly no description line
                description_lines = 0
                break
        else:
            if first_lines:
                description_lines=i
                #description_lines = i
                first_lines = False
            if conf > DESCRIPTION_CONFIDENCE:
                return sample[0:description_lines]
        conf += 1
    return sample[0:description_lines]


HEADER_CONFIDENCE = 1
def guess_headers(sample, empty_columns=None):
    # first store types of first x rows
    types = []
    for i, row in enumerate(sample):
        if i >= HEADER_CONFIDENCE:
            break
        row_types = []
        for c in row:
            t = simple_type_detection.detectType(c)
            row_types.append(t)
        types.append(row_types)

    # now analyse types
    first = types[0]
    if empty_columns:
        first = [i for j, i in enumerate(first) if j not in empty_columns]
    first_is_alpha = all(['ALPHA' in t for t in first])
    if first_is_alpha:
        return sample[0]
    return []

def _detect_header_lines(pending, max_headers=2):

    h_rows = 0
    prev_types = []

    for i, row in enumerate(pending):
        # if we observe more than 3 headers, we assume that our detection failed and the first row is the header
        if i > max_headers:
            h_rows = 1
            break

        log.debug("HEADER? ", row=row)
        row_types = [simple_type_detection.detectType(rowValue) for rowValue in row]

        # if first row contains numeric we assume that there is no header
        if i == 0 and ('FLOAT' in row_types or 'NUM' in row_types or 'NUM+' in row_types):
            h_rows = 0
            break

        if i > 0 and prev_types[i - 1] != set(row_types) and len(row_types) != 0 and len(prev_types[i - 1]) != 0:
            # change of types -> good indicator
            cur_notstrings = not all([x == 'str' or x == 'empty' for x in row_types])
            prev_allstrings = all([x == 'str' or x == 'empty' for x in prev_types[i - 1]])
            if cur_notstrings and prev_allstrings:
                # TODO
                # row type chane
                h_rows = i
                break

        prev_types.append(set(row_types))

    if h_rows < len(pending):
        return pending[:h_rows]
    else:
        # if only strings, assume first line is header
        if len(prev_types) > 0 and all([x == 'str' or x == 'empty' for x in prev_types[0]]):
            return pending[:1]
        return None


def detect_empty_columns(sample):
    empty_col = []
    if len(sample) > 0:
        columns = [ [] for _ in sample[1]]
        for row in sample:
            for j, c in enumerate(row):
                columns[j].append(c)
        for i, col in enumerate(columns):
            if all(['EMPTY' in simple_type_detection.detectType(c) for c in col]):
                empty_col.append(i)
    return empty_col



'''
Created on Feb 1, 2016

@author: jumbrich
@author: svakulenko
'''
import os
import logging

from pyyacp.pyyacp import YACParser
from pyyacp.profiler import profile

SAMPLES_PATH = "sample_csvs/"
SAMPLE_CSVS = ["bevoelkerung.csv", "bezirkszahlen.csv", "hunde_wien.csv", "hunde_wien2.csv"]


def test_parser(path_to_file=SAMPLES_PATH+SAMPLE_CSVS[3]):
    # show debug log info
    logging.basicConfig(level=logging.DEBUG)
    p = YACParser(filename=path_to_file)
    assert p


def test_profiler(path_to_file=SAMPLES_PATH+SAMPLE_CSVS[3], show_counter=False):
    # show debug log info
    logging.basicConfig(level=logging.DEBUG)
    p = YACParser(filename=path_to_file, skip_guess_encoding=False)
    table_info = profile(p)
    assert table_info


def run_profiler(path_to_file=SAMPLES_PATH+SAMPLE_CSVS[3], show_counter=False):
    p = YACParser(filename=path_to_file, skip_guess_encoding=False)
    print 'Header:', p.__dict__['header_line']
    try:
        table_info = profile(p)
        assert table_info
        # show unique values for each of the columns
        for column in table_info['column_unique']:
            print '\n', column
        if show_counter:
            for column in table_info['column_counter']:
                print '\n', column.most_common(30)
    except Exception as e:
        print "Could not process:", path_to_file, e


def collect_columns(path_to_file):
    p = YACParser(filename=path_to_file, skip_guess_encoding=False)
    try:
        table_info = profile(p)
        assert table_info
        return table_info['column_unique']
    except Exception as e:
        print "Could not process:", path_to_file, e


def compare_columns(files=SAMPLE_CSVS, path=SAMPLES_PATH):
    tables = []
    print "Input files:", len(files)
    for sample in files:
        file_path = path + sample
        # test_parser(file_path)
        table_columns = collect_columns(file_path)
        if table_columns:
            tables.append(table_columns)
    # print tables
    print "Tables processed:", len(tables)
    print "Number of columns in each table:"
    print [len(columns) for columns in tables]


if __name__ == '__main__':
    compare_columns()



import functools

from dtpattern.dtpattern1 import pattern_comparator, aggregate_patterns, translate
from pyyacp.profiler import ColumnProfiler, ColumnByCellProfiler, is_not_empty
from pyyacp.timer import timer


class ColumnPatternProfiler(ColumnProfiler,ColumnByCellProfiler):

    def __init__(self, num_patterns=3):
        super(ColumnPatternProfiler, self).__init__('cpp','patterns')
        self.num_patterns=num_patterns
        self.patterns=[]
        self.pa = self.patterns.append

#    @timer(key="cpp_column")
#    def profile_column(self, column, meta):
#        return dtpattern.aggregate(filter(is_not_empty,column), size=self.num_patterns)

    @timer(key="cpp_result")
    def result(self):
        p = sorted(self.patterns, key=functools.cmp_to_key(pattern_comparator))
        #empty
        #reset in case there is another column
        self.patterns = []
        self.pa = self.patterns.append
        return aggregate_patterns(p, size=self.num_patterns)

    def accept(self, cell):
        if is_not_empty(cell):
            self.pa(translate(cell))



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import unittest
import sys
from random import randint, shuffle
from pyyacp.column_format_detector import  aggregate,  aggregate_group_of_same_symbols
from pyyacp.testing.value_pattern_types_list import random_number, random_date, random_time, random_iso8601, \
    random_lowercase_string, random_word, random_words


def for_examples(parameters):

  def tuplify(x):
    if not isinstance(x, tuple):
      return (x,)
    return x

  def decorator(method, parameters=parameters):
    for parameter in (tuplify(x) for x in parameters):

      def method_for_parameter(self, method=method, parameter=parameter):
        method(self, *parameter)
      args_for_parameter = ",".join(repr(v) for v in parameter)
      name_for_parameter = method.__name__ + "(" + args_for_parameter + ")"
      frame = sys._getframe(1)  # pylint: disable-msg=W0212
      frame.f_locals[name_for_parameter] = method_for_parameter
    return None
  return decorator

##basic straight forward aggregation
test_cols_aggregated=[
    (random_number(10,digits=1, fix_len=True), '1'),
    (random_number(10,digits=2, fix_len=True), '11'),
    (random_number(10,digits=3, fix_len=True), '111'),
    (random_number(10,digits=5, fix_len=True), '11111'),
    ( random_number(2,digits=1, fix_len=True)+random_number(2,digits=2, fix_len=True), '01'),
    (random_date(5), '1111-11-11'),
    (random_time(5), '11:11:11'),
    (random_iso8601(5), '1111-11-11C11:11:11'),
    (random_lowercase_string(10, length=5),'ccccc'),
    (random_lowercase_string(5, length=5)+random_lowercase_string(5, length=3),'aaccc'),
    (random_word(10,length=5),'Ccccc'),
    (random_word(5,length=5)+random_word(5,length=3),'Caacc'),
    (random_words(10,length=[5,3]),'Ccccc Ccc'),
    (random_words(5,length=[5,3])+random_words(5,length=[3,5]),'Caacc Caacc'),

]

#This cases are not that straight forward
test_cols_aggregated_complex=[
    ( [u'-1']+random_number(5,digits=1, fix_len=True), '[-]1'), # sometimes numbers contain a minus
    ( [u'-1',u'+1']+random_number(5,digits=1, fix_len=True), '[+/-]1')
]


class TestFormatEncoder(unittest.TestCase):
    def setUp(self):
        self.f = aggregate

    @for_examples(test_cols_aggregated)
    def test_basic(self, x, y):
        res=self.f(values=x)
        self.assertEqual(res[0][0], y)
        self.assertEqual(res[0][1], len(x))

    @for_examples(test_cols_aggregated_complex)
    def test_complex(self, x, y):
        res = self.f(values=x)
        self.assertEqual(res[0][0], y)
        self.assertEqual(res[0][1], len(x))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pyyacp.profiler import  ColumnProfiler


INT='int'
FLOAT='float'
DATETIME='datetime'
DATE='date'
TIME='time'
UNICODE='unicode'

import re
types=[
    (FLOAT,re.compile('^\[?[+/-]*\]?\d+\.\d*$')),
    (INT,re.compile('^\[?[+/-]*\]?\d+$')),
    (DATETIME, re.compile('^\d{4}-\d{2}-\d{2}([C ]+\d{2}:\d{2}:\d{2})$')),
    (DATE, re.compile('^\d{4}-\d{2}-\d{2}$')), #1111-11-11
    (DATE, re.compile('^\d{4}/\d{2}/\d{2}$')), #1111/11/11
    (DATE, re.compile('^\d{2}/\d{2}/\d{4}$')),#11/11/1111

    (TIME, re.compile('^\d{2}:\d{2}(:\d{2})?$')) #1111-11-11
]

class DataTypeDetection(ColumnProfiler):

    def __init__(self):
        super(DataTypeDetection, self).__init__('dtype','data_type')
        self.values=[]

    def profile_column(self, column, meta):
        data_type = UNICODE
        if 'patterns' in meta:
            data_type = self._analyse_ColumnPattern(meta)

        return data_type

    def _analyse_ColumnPattern(self, meta):
        pattern=meta['patterns']
        gtypes = set([])
        for p in pattern:
            for ptype in types:
                if p[0] is not None:
                    m=ptype[1].match(p[0])
                    if m:
                        gtypes.add(ptype[0])
        if len(gtypes) != 1:
            return UNICODE
        d= gtypes.pop()
        if d == INT or d==FLOAT:
            #print meta['stats_min_value']
            if meta['stats_min_value'][0]==u'0':
                #print "LEADING ZERO"
                return UNICODE
        return d


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import re



from pyyacp.profiler.xsd11 import SingleSubType, MultiSubType


decimal_regex=h = re.compile('(\+|-)?([0-9]+(\.[0-9]*)?|\.[0-9]+)$')
class decimal(SingleSubType):

    def __init__(self):
        super(decimal, self).__init__(subType=integer())

    def _isType(self,value):
        return decimal.is_decimal(value)

    @staticmethod
    def is_decimal(value):
        return decimal_regex.match(value)

integer_regex=re.compile('^[\-+]?[0-9]+$')
class integer(MultiSubType):

    def __init__(self):
        subtypes=[
            long()
            ,nonPositiveInteger()
            ,nonNegativeInteger()
        ]
        super(integer, self).__init__(subTypes=subtypes)

    def _isType(self,value):
        return integer.is_integer(value)

    def convert(self, value):
        return int(value)

    @staticmethod
    def is_integer(value):
        return  integer_regex.match(value)

class long(SingleSubType):

    def __init__(self):
        super(long, self).__init__(subType=int())

    def _isType(self,value):
        return long.is_long(value)

    @staticmethod
    def is_long(value):
        return -9223372036854775808<= value <=9223372036854775807

class int(SingleSubType):

    def __init__(self):
        super(int, self).__init__(subType=short())

    def _isType(self,value):
        return int.is_int(value)

    @staticmethod
    def is_int(value):
        return 2147483647>= value >= -2147483648

class short(SingleSubType):

    def __init__(self):
        super(short, self).__init__(subType=byte())

    def _isType(self,value):
        return short.is_short(value)

    @staticmethod
    def is_short(value):
        return 32767>= value >= -32768

class byte(SingleSubType):


    def _isType(self,value):
        return byte.is_byte(value)

    @staticmethod
    def is_byte(value):
        return 127 >= value >= -128





class negativeInteger(SingleSubType):


    def _isType(self,value):
        return negativeInteger.is_negativeInteger(value)

    @staticmethod
    def is_negativeInteger(value):
        return value <= -1

class nonPositiveInteger(SingleSubType):
    def __init__(self):
        super(nonPositiveInteger, self).__init__(subType=negativeInteger())

    def _isType(self, value):
        return nonPositiveInteger.is_nonPositiveInteger(value)

    @staticmethod
    def is_nonPositiveInteger(value):
        return value <= 0


class nonNegativeInteger(MultiSubType):

    def __init__(self):
        subtypes=[
            unsignedLong()
            ,positiveInteger()
        ]
        super(nonNegativeInteger, self).__init__(subTypes=subtypes)


    def _isType(self,value):
        return nonNegativeInteger.is_nonNegInt(value)

    @staticmethod
    def is_nonNegInt(value):
        return value >= 0




class positiveInteger(SingleSubType):


    def _isType(self,value):
        return positiveInteger.is_positiveInteger(value)

    @staticmethod
    def is_positiveInteger(value):
        return value >= 1


class unsignedLong(SingleSubType):

    def __init__(self):
        super(unsignedLong, self).__init__(subType=unsignedInt())

    def _isType(self,value):
        return unsignedLong.is_unsignedLong(value)

    @staticmethod
    def is_unsignedLong(value):
        return 18446744073709551615 >= value >= 0

class unsignedInt(SingleSubType):

    def __init__(self):
        super(unsignedInt, self).__init__(subType=unsignedShort())

    def _isType(self,value):
        return unsignedInt.is_unsignedInt(value)

    @staticmethod
    def is_unsignedInt(value):
        return 4294967295 >= value >= 0

class unsignedShort(SingleSubType):

    def __init__(self):
        super(unsignedShort, self).__init__(subType=unsignedByte())

    def _isType(self,value):
        return unsignedShort.is_unsignedShort(value)

    @staticmethod
    def is_unsignedShort(value):
        return 65535 >= value >= 0

class unsignedByte(SingleSubType):


    def _isType(self,value):
        return unsignedByte.is_unsignedByte(value)

    @staticmethod
    def is_unsignedByte(value):
        return 255 >= value >= 0


# -*- coding: utf-8 -*-
from __future__ import absolute_import
import unittest

from pyyacp.table_structure_helper import AdvanceStructureDetector, SimpleStructureDetector

class TestDescriptionDetectionAdvanceStructureDetector(unittest.TestCase):

    def setUp(self):
        self.structure_detector = AdvanceStructureDetector()
        self.verbose=True


    def test_single_col_no_header(self):
        table=[ ['City'],
                ['Vienna'],
                ['Salzburg']
            ]
        self.assertListEqual([], self.structure_detector.guess_headers(table, verbose=self.verbose))

    def test_single_col_one_header(self):
        table=[ ['Name'],
                ['Tim Tom'],
                ['Max Min']
            ]
        self.assertListEqual([table[0]], self.structure_detector.guess_headers(table, verbose=self.verbose))

    def test_single_col_one_header1(self):
        table=[ ['Vor Nachname'],
                ['Tim Tom'],
                ['Max Min']
            ]
        self.assertListEqual([], self.structure_detector.guess_headers(table, verbose=self.verbose))

    def test_single_col_one_header2(self):
        table=[ ['Count'],
                ['10'],
                ['111']
            ]
        self.assertListEqual([table[0]], self.structure_detector.guess_headers(table, verbose=self.verbose))

    def test_multi_header(self):
        table=[ ['Einwohner'],
                ['Population'],
                ['1799'],
            ]
        self.assertListEqual(table[0:2], self.structure_detector.guess_headers(table, verbose=self.verbose))

    def test_col1_h1_mixed(self):
        table=[ ['Population','City','Country'],
                ['1799', 'Vienna', 'Austria'],
                ['1799', 'Salzburg', 'Austria'],
            ]
        self.assertListEqual(table[0:1], self.structure_detector.guess_headers(table, verbose=self.verbose))





class TestDescriptionDetectionSimpleStructureDetector(TestDescriptionDetectionAdvanceStructureDetector):
    def setUp(self):
        self.structure_detector = SimpleStructureDetector()
        self.verbose = True


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import time
from collections import defaultdict
from itertools import chain, combinations

from pyyacp.profiler import TableProfiler
from pyyacp.timer import Timer, timer

__author__ = 'nina'

class FDProfiler(TableProfiler):

    def __init__(self):
        super(FDProfiler, self).__init__('tane','table_tane')

    @timer(key='fd_table')
    def profile_table(self, datatable):
        fd = TANEAlgorithm()
        stats= fd.process(datatable)
        datatable.meta[self.key]=stats


class TANEAlgorithm(object):
    '''
        Implementation of a TANE algorithm (http://hpi.de/fileadmin/user_upload/fachgebiete/naumann/folien/SS13/DPDC/DPDC_07_FDs.pdf)
        for finding functional dependencies.

        When computing the dependencies, they are described as a tuple of strings (a,b), which means a -> b.
        On the right side, we always have only one column, while on the left side there could be more columns.
        We separate them with '_'. We do it like this, because arrays cannot be keys in a dictionary.
            ('2_6', '7') : 2,6 -> 7  --- columns 2 and 6 infer column 7
            ('3', '4') : 3 -> 4 --- column 3 infers column 4
            ('', '1') : [] -> 1 --- everything infers column 1 = column 1 has always the same value independently of other columns

        When all the dependencies are computed, we transform them to arrays.

        At the end, functional dependency is represented as an array of length >= 1.
            [2,6,7] : 2,6 -> 7  --- columns 2 and 6 infers column 7
            [3,4] : 3 -> 4 --- column 3 infers column 4
            [1] : [] -> 1 --- everything infers column 1 = column 1 has always the same value independently of other columns
    '''

    def process(self, data_table):
        return self.analyse_cols(data_table.columns())

    def analyse_cols(self, columns):
        self.cols=columns
        self.checks = 0
        t0 = time.clock()

        self.relation_column_dict = {}
        self.R = self.getRelations(self.cols) # a set of relations
        self.L = defaultdict(list)
        self.cplus = {'':self.R}
        self.dependencies = []

        self.L[1] = set([x for x in self.R.split('_')])
        l = 1

        while self.L[l] != set([]):
            self.compute_dependencies(l)
            self.prune(l)
            self.L[l+1] = self.generate_next_level(l)
            l += 1


        stats = {}
        stats['dependencies'] = self.dependencies

        t1 = time.clock()
        #stats['fd_processing_time'] = t1 - t0
        #stats['fd_checks'] = self.checks

        stats['suggested_keys'] = self.suggestKeys()
        stats = self.transformToLists(stats)

        return stats['suggested_keys']



    def getRelations(self, cols):
        relations = []
        for i in range(len(cols)):
            relations.append(str(i))
        return '_'.join(relations)


    def compute_dependencies(self, l):
        level = self.L[l]
        for X in level:
            cplus = []

            for a in X.split('_'):
                temp = list(X.split('_'))
                temp.remove(a)
                cplus.append(set( self.getCplus('_'.join(temp)).split('_')))

            self.cplus[X] = '_'.join(sorted(set.intersection(*cplus)))

        for X in level:
            for a in [x for x in X.split('_') if x in self.cplus[X]]:
                temp = list(X.split('_'))
                temp.remove(a)
                dependency = ('_'.join(temp), a)
                if self.checkDependency(dependency):
                    self.dependencies.append(dependency)
                    self.cplus[X] = self.cplus[X].replace(a,'')
                    for r in self.R:
                        if not r in X:
                            self.cplus[X] = self.cplus[X].replace(r,'')


    def prune(self, l):
        level = self.L[l]
        tempLevel = set(level)
        for X in level:
            if self.cplus[X] == '':
                tempLevel.remove(X)

        self.L[l] = set(tempLevel)
        pass


    def generate_next_level(self, i):

        l = set([])

        prefix_blocks = self.prefix_blocks(i)
        for K in prefix_blocks.values():

            if len(K) > 1:
                for c in combinations(K,2):
                    l.add('_'.join(sorted(set(c[0].split('_')+c[1].split('_')))))
        return l


    def prefix_blocks(self, level):
        pb = defaultdict(list)

        for e in self.L[level]:
            prefix = '_'.join(e.split('_')[0:len(e.split('_'))-1])
            pb[prefix].append(e)

        return pb


    def getCplus(self, X):
        if X in self.cplus:
            return self.cplus[X]
        else:
            cplus = set([])
            for a in self.R:
                if X:
                    for b in X:
                        Xtemp = list(X)
                        if a in Xtemp:
                            Xtemp.remove(a)
                        if b in Xtemp:
                            Xtemp.remove(b)
                        dependency = (Xtemp, b)
                        if dependency not in self.dependencies:
                            cplus.add(a)
            return ''.join(sorted(cplus))


    def checkDependency(self, dependency):
        self.checks += 1

        left = dependency[0]
        if len(dependency[0]) > 0:
            left = dependency[0].split('_')

        right = dependency[1]
        rightColumn = self.cols[int(right)]

        leftColumns = [self.cols[int(l)] for l in left]

        values = {}

        for i in range(len(rightColumn)):
            #x is a column
            key= tuple([x[i] for x in leftColumns])
            if key in values:
                if values[key] != rightColumn[i]:
                    return False
            else:
                values[key] = rightColumn[i]

        return True


    def suggestKeys(self):
        '''
            We search for column or combination of columns that defines all the other columns.
            This are key candidates for the table.
        '''

        suggested_keys = []
        dependencies_dict = defaultdict(list)

        # store original dependencies
        for dependency in self.dependencies:
            left = dependency[0]
            right = dependency[1]
            dependencies_dict[left].append(right)


        for dependency in dependencies_dict:
            # check subsets of left hand side of dependency and store what they define
            for sub in map(lambda x: '_'.join(sorted(x)), subsets(dependency)):
                if sub in dependencies_dict:
                    dependencies_dict[dependency] += [a for a in dependencies_dict[sub]]
                    dependencies_dict[dependency] = list(set(dependencies_dict[dependency]))

        dependencies_dict = {k: v for k, v in dependencies_dict.items() if k != ''}

        for d in dependencies_dict:
            temp = self.R.split('_')
            for c in d.split('_'):
                temp.remove(c)
            if set(dependencies_dict[d]) == set(temp):
                suggested_keys.append(d)
        return suggested_keys

    def transformToLists(self, stats):
        '''
            Transform dependencies in '_' form to arrays.
        '''
        suggested_keys = stats['suggested_keys']
        suggested_list = []
        for key in suggested_keys:
            suggested_list.append(list(map(lambda x: int(x),key.split('_'))))

        stats['suggested_keys'] = suggested_list

        FDs = stats['dependencies']
        FD_lists = []
        for fd in FDs:
            if fd[0] != '':
                FD_lists.append(list(map(lambda x:int(x), fd[0].split('_') + [fd[1]])))
            else:
                FD_lists.append(list(map(lambda x:int(x), [fd[1]])))

        stats['dependencies'] = FD_lists
        return stats

def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


def subsets(s):
    return map(set, powerset(s.split('_')))



import random
from math import factorial, log10
import itertools
import math
from math import factorial, log10

def benford_law():
    return [math.log10(1 + 1 / float(i)) * 100.0 for i in xrange(1, 10)]

def pearson(x, y):
    nx = len(x)
    ny = len(y)
    if nx != ny: return 0
    if nx == 0: return 0
    n = float(nx)
    meanx = sum(x) / n
    meany = sum(y) / n
    sdx = math.sqrt(sum([(a - meanx) * (a - meanx) for a in x]) / (n - 1))
    sdy = math.sqrt(sum([(a - meany) * (a - meany) for a in y]) / (n - 1))
    normx = [(a - meanx) / sdx for a in x]
    normy = [(a - meany) / sdy for a in y]
    return sum([normx[i] * normy[i] for i in range(nx)]) / (n - 1)

def getChiSquared(test,expected):
	chi_sq=0
	for t,e in zip(test,expected):
		print('testing :',t)
		print('expected: ',e)
		chi_sq=chi_sq+ ((t-e)*(t-e))/e
	return chi_sq

def passFailChi(chi):
	# degrees of freedom always 8 and assuming 95% confidence
	# from: https://en.wikipedia.org/wiki/Chi-squared_distribution
	# 95% conf: 15.51

	# Null hypothesis H_0: The first digits in the index minute price changes follow the Benford's law
	# Hypothesis H_1: The first digits in the index minute price changes do not follow the Benford's law}
	if (chi<15.51):
		return True
	else:
		return False


def leemis_m(counts, expected):
    '''Based on the Wikipedia article fo Leemis' m.
    I need to have someone check this equation to make sure I got it
    right'''
    deviations = []
    for i in range(len(counts)):
        deviations.append(counts[i] - expected[i])
    m = math.sqrt(sum(counts)) * max(deviations)
    return m


def cho_gaines_distance(counts, expected):
    '''Based on the Wikipedia article for Cho-Gaines' Distance, d.
    I need to have someone check this to make sure it's good.'''
    Sigma = 0
    for i in range(len(counts)):
        Sigma += (counts[i] - expected[i]) ** 2

    distance = math.sqrt(sum(counts) * Sigma)
    return distance


def leemis_critical(m, alpha=.05):
    '''alpha can take on one of the values of .01, .05, .10 for the
    signficance you wish you test. This method can only be run
    after first calculating Leemis' m by running the lemmis_m method'''
    assert alpha == .05 or alpha == .01 or alpha == .10, "This " \
                                                         "module currently supports alpha critical-values of .01, .05, or .10"

    if alpha == .05:
        CV = 0.967
    elif alpha == .01:
        CV = 1.212
    else:  # alpha = .10
        CV = 0.851

    if m > CV:
        print "We can reject the null hypothesis that the data follows " \
              "Benford's law with an m of {} against a crtical value of {}, " \
              "alpha {} using Leemis' m".format(m, CV, alpha)

    else:
        print "We cannot reject the null hypothesis that the data follows " \
              "Benford's law with an m of {} against a crtical value of {}, " \
              "alpha {} using Leemis' m".format(m, CV, alpha)


def cho_gaines_critical(distance, alpha=.05):
    assert alpha == .05 or alpha == .01 or alpha == .10, "This " \
                                                         "module currently supports alpha critical-values of .01, .05, or .10"

    if alpha == .05:
        CV = 1.330
    elif alpha == .01:
        CV = 1.569
    else:  # alpha = .10
        CV = 1.212

    if distance > CV:
        print "We can reject the null hypothesis that the data follows " \
              "Benford's law with an d of {} against a crtical value of {}, " \
              "alpha {} using Cho-Gaines' d".format(distance, CV, alpha)

    else:
        print "We cannot reject the null hypothesis that the data follows " \
              "Benford's law with an d of {} against a crtical value of {}, " \
              "alpha {} using Cho-Gaines' d".format(distance, CV, alpha)


def get_leading_number(cell):
    numbers = "123456789"
    l = map(int, filter(lambda x: x in numbers, cell))
    return l[0] if len(l)>0 else 0

def getExpectedBenfordCount(size):
	return map(lambda x:x*0.01*size, benford_law())


from faker import Factory
fake = Factory.create()
#r=[unicode(fake.random_digit()*0.5) for i in range (1,100000)]
r=[unicode(random.randint(0, 10000)) for a in range(1000)]





#r=[unicode(factorial(i)) for i in range(0,100)]
#print r
print r

#get leadign numpers
x=sorted(map(get_leading_number,r))
import numpy as np
d=np.array([x.count(k) for k in range(1,10)])

#b=[math.log10(1 + 1 / float(i)) * 100.0 for i in xrange(1, 10)]
#d = np.array(getExpectedBenfordCount(100))


print d
d_prob= d/float(sum(d))*100
print d_prob

##
y= getExpectedBenfordCount(len(x))
print 'expected'
print y
y_prob= benford_law()
print y_prob


def chisq_stat(O, E):
    return sum([(o - e) ** 2 / e for (o, e) in zip(O, E)])


chi=getChiSquared(d,y)
print 'CHI', chi, passFailChi(chi)
print 'CHI', chisq_stat(d,y)

m=leemis_m(d_prob,y_prob )
print 'LEEMIS',m, leemis_critical(m)
dist=cho_gaines_distance(d_prob,y_prob)
print "CHO", dist, cho_gaines_critical(dist)



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import traceback

import sys

from pyyacp.html.to_html import to_html
import  pyyacp.datatable as datatable
from pyyacp.profiler import ColumnRegexProfiler, ColumnByCellProfilerSet, \
    ColumnProfiler, ColumnProfilerSet
from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from profiler.profiling import apply_profilers

from pyyacp.profiler.column_stats_profiler import ColumnStatsProfiler
from pyyacp.profiler.data_type_detection import DataTypeDetection
from pyyacp.profiler.data_type_interpretation import DataTypeInterpretation
from pyyacp.profiler.distributions import CharacterDistributionProfiler, BenfordsLawDistribution
from pyyacp.profiler.fdprofiler import FDProfiler
from pyyacp.profiler.xsd11_type_detection import XSDTypeDetection
from pyyacp.table_structure_helper import AdvanceStructureDetector
from pyyacp.testing.csv_iterator import csvContent_iter

SAMPLES_PATH = "./sample_csvs"
from os import listdir
from os.path import isfile, join
onlyfiles = [join(SAMPLES_PATH, f) for f in listdir(SAMPLES_PATH) if isfile(join(SAMPLES_PATH, f))]




def from_csv_iter(portalID='data_wu_ac_at', snapshot=None):
    profilers=[ColumnPatternProfiler(),DataTypeDetection(), ColumnStatsProfiler()]#FDProfiler(), ,ColumnStatsProfiler(), DataTypeDetection()]#,XSDTypeDetection()] #,ColumnRegexProfiler()]#,XSDTypeDetection()],,ColumnStatsProfiler()
    cnt=0
    for uri, csv_file in csvContent_iter(portalID, snapshot=snapshot):
        cnt+=1
        print("{}, {} -> {}".format(cnt,uri, csv_file))
        try:

            from pyyacp.pyyacp import YACParser
            yacp = YACParser(filename=csv_file,sample_size=1800, structure_detector = AdvanceStructureDetector())
            table=datatable.parseDataTables(yacp, url=uri)
            profilers=[ColumnByCellProfilerSet([ColumnPatternProfiler, ColumnStatsProfiler, CharacterDistributionProfiler, BenfordsLawDistribution]) , ColumnProfilerSet([DataTypeDetection,DataTypeInterpretation])]
            apply_profilers(table,profilers=profilers)



            to_html(table, cnt, dir='.')

        except Exception as e:
            print(traceback.format_exc())
            print(sys.exc_info()[0])
            print(e)
        #print'next'
        if cnt>10:
            break

if __name__ == '__main__':
    from_csv_iter(snapshot=1803)



from collections import defaultdict

from pyyacp.profiling import HeaderRowProfiler
from pyyacp.utils.timer import Timer
# nltk tools
from nltk.corpus import wordnet
from nltk.stem.snowball import SnowballStemmer

import re


class HeaderTokensProfiler(HeaderRowProfiler):
    def __init__(self, id='header_tokens'):
        super(HeaderTokensProfiler, self).__init__(id)
        self.key='header_tokens'
        self.tokens = defaultdict(int)

    def analyse_HRow(self, header_row):
        for header_cell in header_row.cells:
            header = header_cell.value

            if len(header.split(' ')) > 1:
                # multiple words
                self.tokens['MULTIPLE'] += 1
            elif len(header.split('_')) > 1:
                # underscore separated
                self.tokens['UNDERSCORE'] += 1
                pass
            elif len(re.sub("([a-z])([A-Z])","\g<1> \g<2>", header).split(' ')) > 1:
                # camelcase separated
                self.tokens['CAMELCASE'] += 1
                pass
            else:
                # single word
                self.tokens['SINGLE'] += 1

    def getResult(self):
        return dict(self.tokens)

class HeaderReadabilityProfiler(HeaderRowProfiler):

    def __init__(self, id='header_readability'):
        super(HeaderReadabilityProfiler, self).__init__(id)
        self.key='header_readability'
        self.languages = ['english', 'german']
        self.stemmer = {lan: SnowballStemmer(lan) for lan in self.languages}
        self.header = defaultdict(int)

    def analyse_HRow(self, header_row):
        with Timer(key="HeaderReadabilityProfiler") as t:
            for header_cell in header_row.cells:
                self.header['TOTAL'] += 1
                for lan in self.languages:
                    header = header_cell.value
                    # check if word (or stemmed word) is in wordnet knowledge base
                    stemmed_h = self.stemmer[lan].stem(header)
                    if wordnet.synsets(header) or wordnet.synsets(stemmed_h):
                        self.header['READABLE'] += 1
                        break
                        #self.header[header][lan] = True
                    #else:
                        #self.header[header][lan] = False

    def getResult(self):
        return dict(self.header)


from profiler.fdprofiler import TANEAlgorithm

tane= TANEAlgorithm()
size=10000
colNo=20
def f(i,k):
    if k==2:
        return 2
    else:
        return i*k
cols={k:[f(i,k) for i in range(0,size)] for k in range(1,colNo+1)}
import time
s=time.time()
print tane.analyse_cols(cols.values())
print (time.time()-s)*1000


"""------------------------------------------------------------------------------------------
TANE Algorithm for discovery of exact functional dependencies
Author: Nabiha Asghar, nasghar@uwaterloo.ca
February 2015
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
----------------------------------------------------------------------------------------------"""
from pandas import *
import pandas as pd
from collections import defaultdict
import numpy as NP
import sys

from pyyacp.timer import Timer


def list_duplicates(seq):
    tally = defaultdict(list)
    for i, item in enumerate(seq):
        tally[item].append(i)
    return ((key, locs) for key, locs in tally.items()
            if len(locs) > 0)


def findCplus(x):  # this computes the Cplus of x as an intersection of smaller Cplus sets
    global dictCplus
    thesets = []
    for a in x:
        if x.replace(a, '') in dictCplus.keys():
            temp = dictCplus[x.replace(a, '')]
        else:
            temp = findCplus(x.replace(a, ''))  # compute C+(X\{A}) for each A at a time
        # dictCplus[x.replace(a,'')] = temp
        thesets.insert(0, set(temp))
    if list(set.intersection(*thesets)) == []:
        cplus = []
    else:
        cplus = list(set.intersection(*thesets))  # compute the intersection in line 2 of pseudocode
    return cplus


def compute_dependencies(level, listofcols):
    global dictCplus
    global finallistofFDs
    global listofcolumns
    for x in level:
        thesets = []
        for a in x:
            if x.replace(a, '') in dictCplus.keys():
                temp = dictCplus[x.replace(a, '')]
            else:
                temp = computeCplus(x.replace(a, ''))  # compute C+(X\{A}) for each A at a time
                dictCplus[x.replace(a, '')] = temp
            thesets.insert(0, set(temp))
        if list(set.intersection(*thesets)) == []:
            dictCplus[x] = []
        else:
            dictCplus[x] = list(set.intersection(*thesets))  # compute the intersection in line 2 of pseudocode
    for x in level:
        for a in x:
            if a in dictCplus[x]:
                # if x=='BCJ': print "dictCplus['BCJ'] = ", dictCplus[x]
                if validfd(x.replace(a, ''), a):  # line 5
                    finallistofFDs.append([x.replace(a, ''), a])  # line 6
                    dictCplus[x].remove(a)  # line 7

                    listofcols = listofcolumns[:]
                    for j in x:  # this loop computes R\X
                        if j in listofcols: listofcols.remove(j)

                    for b in listofcols:  # this loop removes each b in R\X from C+(X)
                        if b in dictCplus[x]: dictCplus[x].remove(b)


def computeCplus(
        x):  # this computes the Cplus from the first definition in section 3.2.2 of TANE paper. output should be a list of single attributes
    global listofcolumns
    listofcols = listofcolumns[:]
    if x == '': return listofcols  # because C+{phi} = R
    cplus = []
    for a in listofcols:
        for b in x:
            temp = x.replace(a, '')
            temp = temp.replace(b, '')
            if not validfd(temp, b):
                cplus.append(a)
    return cplus


def validfd(y, z):
    if y == '' or z == '': return False
    ey = computeE(y)
    eyz = computeE(y + z)
    if ey == eyz:
        return True
    else:
        return False


def computeE(x):
    global totaltuples
    global dictpartitions
    doublenorm = 0
    for i in dictpartitions[''.join(sorted(x))]:
        doublenorm = doublenorm + len(i)
    e = (doublenorm - len(dictpartitions[''.join(sorted(x))])) / float(totaltuples)
    return e


def check_superkey(x):
    global dictpartitions
    if ((dictpartitions[x] == [[]]) or (dictpartitions[x] == [])):
        return True
    else:
        return False


def prune(level):
    global dictCplus
    global finallistofFDs
    stufftobedeletedfromlevel = []
    for x in level:  # line 1
        if dictCplus[x] == []:  # line 2
            level.remove(x)  # line 3
        if check_superkey(x):  # line 4   ### should this check for a key, instead of super key??? Not sure.
            temp = dictCplus[x][:]
            for i in x:  # this loop computes C+(X) \ X
                if i in temp: temp.remove(i)
            for a in temp:  # line 5
                thesets = []
                for b in x:
                    if not (''.join(sorted((x + a).replace(b, ''))) in dictCplus.keys()):
                        dictCplus[''.join(sorted((x + a).replace(b, '')))] = findCplus(
                            ''.join(sorted((x + a).replace(b, ''))))
                    thesets.insert(0, set(dictCplus[''.join(sorted((x + a).replace(b, '')))]))
                if a in list(set.intersection(*thesets)):  # line 6
                    finallistofFDs.append([x, a])  # line 7
                    # print "adding key FD: ", [x,a]
            if x in level: stufftobedeletedfromlevel.append(x)  # line 8
    for item in stufftobedeletedfromlevel:
        level.remove(item)


def generate_next_level(level):
    nextlevel = []
    for i in range(0, len(level)):  # pick an element
        for j in range(i + 1, len(level)):  # compare it to every element that comes after it.
            if ((not level[i] == level[j]) and level[i][0:-1] == level[j][0:-1]):  # i.e. line 2 and 3
                x = level[i] + level[j][-1]  # line 4
                flag = True
                for a in x:  # this entire for loop is for the 'for all' check in line 5
                    if not (x.replace(a, '') in level):
                        flag = False
                if flag == True:
                    nextlevel.append(x)
                    stripped_product(x, level[i], level[
                        j])  # compute partition of x as pi_y * pi_z (where y is level[i] and z is level[j])
    return nextlevel


def stripped_product(x, y, z):
    global dictpartitions
    global tableT
    tableS = [''] * len(tableT)
    partitionY = dictpartitions[''.join(sorted(y))]  # partitionY is a list of lists, each list is an equivalence class
    partitionZ = dictpartitions[''.join(sorted(z))]
    partitionofx = []  # line 1
    for i in range(len(partitionY)):  # line 2
        for t in partitionY[i]:  # line 3
            tableT[t] = i
        tableS[i] = ''  # line 4
    for i in range(len(partitionZ)):  # line 5
        for t in partitionZ[i]:  # line 6
            if (not (tableT[t] == 'NULL')):  # line 7
                tableS[tableT[t]] = sorted(list(set(tableS[tableT[t]]) | set([t])))
        for t in partitionZ[i]:  # line 8
            if (not (tableT[t] == 'NULL')) and len(tableS[tableT[t]]) >= 2:  # line 9
                partitionofx.append(tableS[tableT[t]])
            if not (tableT[t] == 'NULL'): tableS[tableT[t]] = ''  # line 10
    for i in range(len(partitionY)):  # line 11
        for t in partitionY[i]:  # line 12
            tableT[t] = 'NULL'
    dictpartitions[''.join(sorted(x))] = partitionofx


def computeSingletonPartitions(listofcols):
    global data2D
    global dictpartitions
    for a in listofcols:
        dictpartitions[a] = []
        for element in list_duplicates(data2D[
                                           a].tolist()):  # list_duplicates returns 2-tuples, where 1st is a value, and 2nd is a list of indices where that value occurs
            if len(element[1]) > 1:  # ignore singleton equivalence classes
                dictpartitions[a].append(element[1])



with Timer(verbose=True):
    cols={'A':[i for i in range(0,100)],
          'B':[i%2 for i in range(0,100)],
          'C':[i%3 for i in range(0,100)],
          'D': [i % 4 for i in range(0, 100)],
          'E': [i % 2 for i in range(0, 100)]
          }
    data2D=pd.DataFrame(cols)

    totaltuples = len(data2D.index)
    listofcolumns = list(data2D.columns.values) # returns ['A', 'B', 'C', 'D', .....]

    tableT = ['NULL']*totaltuples # this is for the table T used in the function stripped_product

    L0 = []
    dictCplus = {'NULL': listofcolumns}
    dictpartitions = {} # maps 'stringslikethis' to a list of lists, each of which contains indices
    computeSingletonPartitions(listofcolumns)
    finallistofFDs=[]
    #print dictCplus['NULL']
    L1=listofcolumns[:]  # L1 is a copy of listofcolumns
    l=1

    L = [L0,L1]

    while (not (L[l] == [])):
        compute_dependencies(L[l],listofcolumns[:])
        prune(L[l])
        temp = generate_next_level(L[l])
        L.append(temp)
        l=l+1

    print "List of all FDs: " , finallistofFDs
    print "Total number of FDs found: ", len(finallistofFDs)




# -*- coding: utf-8 -*-
import codecs
import unittest
import string
import sys

from pyyacp.column_format_detector import translate


def for_examples(parameters):

  def tuplify(x):
    if not isinstance(x, tuple):
      return (x,)
    return x

  def decorator(method, parameters=parameters):
    for parameter in (tuplify(x) for x in parameters):

      def method_for_parameter(self, method=method, parameter=parameter):
        method(self, *parameter)
      args_for_parameter = ",".join(repr(v) for v in parameter)
      name_for_parameter = method.__name__ + "(" + args_for_parameter + ")"
      frame = sys._getframe(1)  # pylint: disable-msg=W0212
      frame.f_locals[name_for_parameter] = method_for_parameter
    return None
  return decorator


test_cases=[ (c,'1') for c in string.digits]
test_cases.extend([ (c,'a') for c in string.ascii_lowercase])
test_cases.extend([ (c,'A') for c in string.ascii_uppercase])

encoding=[('ä','a'),('ö','a'),('ü','a'),
          ('Ä', 'A'), ('Ö', 'A'), ('Ü', 'A')]

special=[('?','?')]

mix=[('Wien 1110', 'Aaaa 1111')]

class TestFormatEncoder(unittest.TestCase):

    @for_examples(test_cases)
    def test_basic(self, x, y):
        self.assertEqual(translate(x), y)

    @for_examples(encoding)
    def test_encoding(self, x, y):
        self.assertEqual(translate(x), y)

    @for_examples(special)
    def test_special(self, x, y):
        self.assertEqual(translate(x), y)

    @for_examples(mix)
    def test_mix(self, x, y):
        self.assertEqual(translate(x), y)


if __name__ == '__main__':
    unittest.main()


#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import traceback

import sys

import  pyyacp.datatable as datatable

from pyyacp.table_structure_helper import AdvanceStructureDetector


#filename=csv_file
from pyyacp import YACParser

from pyyacp.profiler.profiling import apply_profilers

url='http://nsandi-corporate.com/wp-content/uploads/2015/02/transparency-25k-08-2013.csv'
url='http://www.win2day.at/download/lo_1992.csv'
url='http://data.wu.ac.at/data/unibib/diff/JREK/2015-07-07.csv'
url='http://www.win2day.at/download/etw_2006.csv'
url='http://www.win2day.at/download/jo_2004.csv'
url='http://www.win2day.at/download/jo_1992.csv'
url='http://www.wien.gv.at/politik/wahlen/ogd/bv151_99999999_9999_spr.csv'


url='http://www.win2day.at/download/tw_2002.csv'
url='http://wko.at/statistik/opendata/sm/OGD_mgstat_sm_s11-1_bld_sp4.csv'
url='http://wahlen.tirol.gv.at/gemeinderatswahl_2010/dokumente/wahl25.csv'
url='http://www.win2day.at/download/etw_2001.csv'
url='http://www.wien.gv.at/statistik/ogd/vie_106.csv'
#url='http://data.wu.ac.at/portal/dataset/fdb16224-5f6c-482b-932f-e5fe12f52991/resource/a545bb37-0563-4312-be3f-b36a793c0764/download/allcoursesandorgid14s.csv'




structure_detector = AdvanceStructureDetector()





def profile(url, filename,max_tables=1, sample_size=1800):
    yacp = YACParser(filename=filename, url=url, structure_detector=structure_detector, sample_size=sample_size)
    if url is None:
        url='http://example.org/table'
    tables = datatable.parseDataTables(yacp, url=url, max_tables=max_tables)

    if max_tables==1:
        table=tables


    apply_profilers(table)


    table.print_summary()


url=None
url="http://data.statistik.gv.at/data/OGD_bpihbaugg2015_BPI_H2015_1_C-A10-0.csv"
filename=None
#filename="/Users/jumbrich/data/csv_catalog_new/data/8/1906eab434e57878a013fbd4622257cc775375a8/856294b5a0f32a4d941843c7dadac1ad.gz"
profile(url=url,filename=filename)



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = '0.3'






"""Main module."""
import csv

import anycsv

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from pyyacp.table_structure_helper import SimpleStructureDetector

import structlog
log = structlog.get_logger()



class YACParserException(Exception):
    pass


YACParser_config={
    'filename':None,
    'url':None,
    'content':None,
    'sample_size': 20,
    'skip_guess_encoding':False,
    'structure_detector':SimpleStructureDetector(),
    'max_file_size':-1
}

class YACParser(object):
    def __init__(self, filename=None, url=None, content=None, sample_size=20, skip_guess_encoding=False, structure_detector=SimpleStructureDetector(), max_file_size=-1):

        self.table = anycsv.reader(filename=filename, url=url, content=content, skip_guess_encoding=skip_guess_encoding, sniff_lines=sample_size,max_file_size=max_file_size)

        keys = ['encoding', 'url', 'filename', 'delimiter', 'quotechar','lineterminator','skipinitialspace','quoting','delimiter','doublequote']
        self.meta = {}
        for k,v in self.table.__dict__.items():
            if k in keys and v:
                self.meta[k] = v
        for k,v in self.table.dialect.items():
            if k in keys:
                self.meta[k] = v
        if 'url' in self.meta:
            self.meta['uri'] = self.meta.pop('url')

        #self.meta['dialect'] = self.table.dialect
        log.debug("Input file dialect", dialect=self.table.dialect, encoding=self.meta['encoding'])
        self.sample = []
        for i, row in enumerate(self.table):
            if i >= sample_size:
                break
            self.sample.append(row)

        # i would include emtpy lines for now
        self.descriptionLines = structure_detector.guess_description_lines(self.sample)

        if self.descriptionLines is None:
            raise ValueError("structure_detector should return a value, if no description lines exists, return empty")
        # allow multple header lines if existing, if none exist return 0
        self.header_lines = structure_detector.guess_headers(self.sample)
        if self.header_lines is None:
            raise ValueError("structure_detector should return a value, if no header lines exists, return empty")
        self.columns=structure_detector.guess_columns(self.sample)

        self.table.seek_line(0)#len(self.descriptionLines) + len(self.header_lines))

    def seek_line(self, lineNo):
        self.table.seek_line(lineNo)


    def __next__(self):
        return self.table.next()

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self


    def getHeader(self):
        if self.header_line:
            return self.header_line
        else:
            return ['col'+str(i) for i in range(self.columns)]

    def reset(self):
        if self.header_line:
            self.table.seek_line(self.description + 1)
        else:
            self.table.seek_line(self.description)

    def generate(self, delimiter=',', newline='\n', header=True, comments=True):
        """
        :param delimiter: The delimiter symbol. The default is ",".
        :param newline: The line separator. The default is "\n".
        :param header: There will be a header in the output (if no header detected then col0,col1, ...)
        :param commentPrefix: The prefix for comments in the first rows. If false, any comments will be ignored.
        :return: A string representation of the CSV table
        """
        csvstream = StringIO()
        w = csv.writer(csvstream, encoding="utf-8", delimiter=delimiter, lineterminator=newline)

        # write description lines at top
        if comments:
            for line in self.descriptionLines:
                w.writerow(line)
        if header:
            w.writerow(self.getHeader())
        for row in self:
            w.writerow(row)
        return csvstream.getvalue()



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import traceback

import sys

import  pyyacp.datatable as datatable
from pyyacp import YACParser
from pyyacp.profiler.profiling import apply_profilers
from pyyacp.table_structure_helper import AdvanceStructureDetector

from pyyacp.profiler.fdprofiler import FDProfiler
from pyyacp.profiler import ColumnProfilerSet, ColumnByCellProfilerSet
from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from pyyacp.profiler.column_stats_profiler import ColumnStatsProfiler
from pyyacp.profiler.data_type_detection import DataTypeDetection
from pyyacp.profiler.cset_detection import CSetDetection
from pyyacp.profiler.data_type_interpretation import DataTypeInterpretation
from pyyacp.profiler.distributions import CharacterDistributionProfiler, BenfordsLawDistribution
from pyyacp.table_structure_helper import AdvanceStructureDetector


structure_detector=AdvanceStructureDetector


default_profilers = [FDProfiler, ColumnByCellProfilerSet(
            [ColumnPatternProfiler, ColumnStatsProfiler, CharacterDistributionProfiler, BenfordsLawDistribution]),
                             ColumnProfilerSet([DataTypeDetection, DataTypeInterpretation,CSetDetection])]






def profile_csv(csv_file):
    try:



        # filename=csv_file

        yacp = YACParser(filename=csv_file, sample_size=1800, structure_detector=AdvanceStructureDetector())
        table = datatable.parseDataTables(yacp, url="http://example.org/")

        apply_profilers(table, profilers=default_profilers)


    except Exception as e:
        print(traceback.format_exc())
        print(sys.exc_info()[0])
        print(e)


csv_file="/Users/jumbrich/data/mimesis_csvs/dummy/de-at_address_r100xc20.csv"
profile_csv(csv_file)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
__author__ = 'nina'

import re
import dateutil.parser


def detectType(value):

    num = '^([0-9]*)$'
    numplus = '^([0-9!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c]*)$'
    plus = '^([!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c]*)$'
    alphanum = '^([^!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c]*)$'
    alpha = '^([^0-9!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c]*)$'
    alphaplus = '^([^0-9]*)$'
    float =  '^([0-9]*[.,][0-9]*)$'


    if len(value) == 0:
        return "EMPTY"
    elif re.match(float, value):
        return "FLOAT"
    elif re.match(plus, value):
        return "+"
    elif re.search(num, value):
        return "NUM"
    elif re.search(numplus, value):
        try:
            dateutil.parser.parse(value)
            return "TEMP"
        except:
            return "NUM+"
    elif re.search(alpha, value):
        return "ALPHA"
    elif re.search(alphaplus, value):
        return "ALPHA+"
    elif re.search(alphanum, value):
        return "ALPHANUM"
    else:
        try:
            dateutil.parser.parse(value)
            return "TEMP"
        except:
            return "ALPHANUM+"



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import re

from pyyacp.profiler.xsd11 import MultiSubType,SingleSubType
from pyyacp.profiler.xsd11.anyatomic_decimal import decimal

class anySimpleType(MultiSubType):

    def __init__(self):
        subTypes=[
            anyAtomicType()
        ]
        super(anySimpleType, self).__init__(subTypes=subTypes)

    def _isType(self, value):
        return True


class anyAtomicType(MultiSubType):

    def __init__(self):
        subTypes=[
            boolean()
            ,anyURI()
            ,base64Binary()
            ,date()
            ,dateTime()
            ,decimal()
            ,double()
            ,duration()
            ,float()
            ,gDay()
            ,gMonthDay()
            ,gMonth()
            ,gYear()
            ,gYearMonth()
            ,hexBinary()
            ,time()
            ,string()
        ]
        super(anyAtomicType, self).__init__(subTypes=subTypes)

    def _isType(self, value):
        return True

###############
# anyAtomicType
################
class boolean(SingleSubType):

    def _isType(self,value):
        return value in ['true','false','1','0']


class anyURI(SingleSubType):

    def _isType(self,value):
        return anyURI.is_iri(value)

    @staticmethod
    def is_iri(value):
        try:
            import rfc3987
            rfc3987.parse(value, rule="URI")
            return True
        except Exception as e:
            return False



class base64Binary(SingleSubType):

    def _isType(self,value):
        return base64Binary.is_base64binary(value)

    @staticmethod
    def is_base64binary(value):
        import base64
        try:
            base64.decodestring(value)
            return True
        except Exception:
            return False



date_regex=re.compile('^-?([1-9][0-9]{3,}|0[0-9]{3})-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$')
class date(SingleSubType):


    def _isType(self, value):
        return date.is_date(value)
    @staticmethod
    def is_date(value):
        return date_regex.search(value)

class dateTime(SingleSubType):

    def __init__(self):
        super(dateTime, self).__init__(subType=dateTimeStamp())

    def _isType(self,value):
        return dateTime.is_datetime(value)

    @staticmethod
    def is_datetime(value):
        regex='^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
        if re.search(regex, value):
            return True
        return False



class dateTimeStamp(SingleSubType):

    def _isType(self,value):
        return dateTimeStamp.is_datetimestamp(value)

    @staticmethod
    def is_datetimestamp(value):
        regex='^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)\.*(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
        if re.search(regex, value):
            return True
        return False


class double(SingleSubType):

    def _isType(self,value):
        return double.is_double(value)

    @staticmethod
    def is_double(value):
        regex='^((\+|-)?([0-9]+(\.[0-9]*)?|\.[0-9]+)([Ee](\+|-)?[0-9]+)?|(\+|-)?INF|NaN)$'
        if re.match(regex, value):
            return True
        return False

class duration(MultiSubType):

    def __init__(self):
        subTypes=[
            yearMonthDuration()
            ,dayTimeDuration()

        ]
        super(duration, self).__init__(subTypes=subTypes)

    def _isType(self,value):
        return duration.is_duration(value)

    @staticmethod
    def is_duration(value):
        regex='-?P((([0-9]+Y([0-9]+M)?([0-9]+D)?|([0-9]+M)([0-9]+D)?|([0-9]+D))(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S)))?)|(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S))))'
        if re.match(regex, value):
            return True
        return False


class yearMonthDuration(SingleSubType):
    def _isType(self,value):
        return yearMonthDuration.is_yearMonthDuration(value)

    @staticmethod
    def is_yearMonthDuration(value):
        regex='^(-?P((([0-9]+Y)([0-9]+M)?)|([0-9]+M)))$'
        if re.match(regex, value):
            return True
        return False

class dayTimeDuration(SingleSubType):
    def _isType(self,value):
        return dayTimeDuration.is_dayTimeDuration(value)

    @staticmethod
    def is_dayTimeDuration(value):
        regex='^([^YM]*[DT].*)$'
        if re.match(regex, value):
            return True
        return False



class float(SingleSubType):

    def _isType(self,value):
        return float.is_float(value)

    @staticmethod
    def is_float(value):
        regex='^((\+|-)?([0-9]+(\.[0-9]*)?|\.[0-9]+)([Ee](\+|-)?[0-9]+)?|(\+|-)?INF|NaN)$'
        if re.match(regex, value):
            return True
        return False

class gDay(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(---(0[1-9]|[12][0-9]|3[01])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False



class gMonthDay(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(--(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False


class gMonth(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(--(0[1-9]|1[0-2])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False


class gYear(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(-?([1-9][0-9]{3,}|0[0-9]{3})(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False


class gYearMonth(SingleSubType):

    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(-?([1-9][0-9]{3,}|0[0-9]{3})-(0[1-9]|1[0-2])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False



class hexBinary(SingleSubType):

    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(([0-9a-fA-F]{2})*)$'
        if re.match(regex, value):
            return True
        return False


class time(SingleSubType):

    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^((([01][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9](\.[0-9]+)?|(24:00:00(\.0+)?))(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?)$'
        if re.match(regex, value):
            return True
        return False



class string(SingleSubType):

    def __init__(self):
        super(string, self).__init__(subType=dateTimeStamp())

    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^(([0-9a-fA-F]{2})*)$'
        if re.match(regex, value):
            return True
        return False





class language(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^([a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*)$'
        if re.match(regex, value):
            return True
        return False


class NMTOKEN(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^([-._:A-Za-z0-9]+)$'
        if re.match(regex, value):
            return True
        return False





class name(SingleSubType):


    def _isType(self,value):
        return self.__class__.is_valid(value)

    @staticmethod
    def is_valid(value):
        regex='^([_:A-Za-z][-._:A-Za-z0-9]*)$'
        if re.match(regex, value):
            return True
        return False



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from dtpattern.dtpattern1 import translate_all, l1_aggregate, l2_aggregate, patterns_to_unique_symbols
from pyyacp.profiler import ColumnProfiler
from pyyacp.profiler.data_type_detection import DATETIME, INT, UNICODE

import re
SENT=re.compile('^(C[c])+( [Cc]+)*[\.!?]?$')
ENT=re.compile('^(C[c])+( [Cc]+){2,3}$')

class CSetDetection(ColumnProfiler):

    def __init__(self):
        super(CSetDetection, self).__init__('cset','cset')


    def profile_column(self, column, meta):

        return self._analyse_ColumnPattern(column, meta)


    def _analyse_ColumnPattern(self, column,meta):
        patterns = translate_all(column)

        res=set([])
        for p in patterns:
            s=set(p)
            res.update(s)



        return ''.join(sorted(list(set(res))))





#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import structlog

from pyyacp.profiler.empty_cell_detection import is_not_empty

log = structlog.get_logger()


class ColumnProfilerSet(object):
    """
    This class provides a wrapper for profilers which take a column as input
    """
    def __init__(self, profilers=[]):
        self.profiler_factories=profilers

    def _init(self):
        self.profilers=[p() for p in self.profiler_factories]

    def profile_column(self, column, meta):
        self._init()
        for p in self.profilers:
            result = p.profile_column(column, meta)
            self._add_results(meta, result, p)

    def _add_results(self, meta, results, p):
        if isinstance(results, dict):
            for k, v in results.items():
                meta["{}_{}".format(p.key, k)] = v
        else:
            meta["{}".format(p.key)] = results


class ColumnByCellProfilerSet(ColumnProfilerSet):

    def __init__(self, profilers=[]):
        super(ColumnByCellProfilerSet, self).__init__(profilers=profilers)

    def profile_column_by_cell(self, column, meta):
        self._init()
        pa=[p.accept for p in self.profilers]

        for cell in column:
            for p in pa:
                p(cell)
        for p in self.profilers:
            result = p.result()
            self._add_results(meta, result, p)

    def profile_column(self, column, meta):
        self.profile_column_by_cell(column, meta)



class Profiler(object):
    def __init__(self, id, key):
        self.id=id
        self.key=key

class TableProfiler(Profiler):
    def __init__(self, id, key):
        super(TableProfiler, self).__init__(id,key)

    def profile_table(self, table):
        pass

class ColumnProfiler(Profiler):
    def __init__(self, id, key):
        super(ColumnProfiler, self).__init__(id,key)

    def profile_column(self, column, meta):
        pass

class ColumnByCellProfiler(Profiler):
    def __init__(self, id, key):
        super(ColumnByCellProfiler, self).__init__(id,key)

    def result(self):
        pass

    def accept(self, cell):
        pass


class ColumnRegexProfiler(Profiler):
    def __init__(self):
        super(ColumnRegexProfiler, self).__init__('crp','col_regex')
        self.values=[]

    def profile_column(self, column, meta):
        return dtpattern.aggregate(column, size=self.num_patterns)

    def result(self):
        res=self.profile_column(self.values, None)
        self.values=[]
        return res

    def accept(self, cell):
        self.values.append(cell)

    def _profile(self, table):
        for i, col in enumerate(table.columns()):
            try:
                regex = regroup.DAWG.from_iter(col).serialize()
                table.column_metadata[i][self.key]=regex
            except Exception as e:
                table.column_metadata[i][self.key]="exc:{}".format(str(e.__class__))








#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import defaultdict

from pyyacp.profiler import Profiler
from pyyacp.profiler.xsd11.types import anySimpleType
from pyyacp.timer import Timer


class XSDTypeDetection(Profiler):

    def __init__(self):
        super(XSDTypeDetection, self).__init__('xsd','col_xsd')
        self.a = anySimpleType()


    def parseTypes(self, type, d, i=0):
        d1 = d.setdefault(type.type(), {'c': 0})
        #print "{}{} {}".format(' '*i,i,type.type())
        if hasattr(type,'_subTypes'):
            d1['st']={}
            st=type._subTypes
            for s in st:
                self.parseTypes(s, d1['st'],i=i+2)

        if hasattr(type, '_subType'):
            st = type._subType
            if st is not None:
                d1['st'] = {}
                self.parseTypes(st, d1['st'], i=i + 2)

    def _profile(self, table):
        for i, col in enumerate(table.columns()):
            patterns = self._analyse_Column(col)
            table.column_metadata[i][self.key] = patterns

    def _analyse_Column(self,col):
        c = {}
        self.parseTypes(self.a, c)

        counts = defaultdict(int)
        for value in col:
            xsdtype = self.a.detectType(value)
            #self.update_Counts(xsdtype, counts)
            for d in xsdtype:
                counts[d] += 1

        def update(c,counts):
            for k,v in c.items():
                if k in counts:
                    v['c']=counts[k]
                    if 'st' in v:
                        update(v['st'], counts)
        update(c,counts)
        def cleanup(c):
            for k,v in c.items():
                #print k,v
                if v['c']==0:
                #    print "del",k
                    del c[k]
                elif 'st' in v:
                    cleanup(v['st'])
                    if len(v['st'])==0:
                        del v['st']
        cleanup(c)
        return c


    def update_Counts(self, types, counts):
        if isinstance(types, dict):
            for k,v in types.items():
                if v is True:
                    counts=counts.setdefault(k,{'c':0})
                    counts['c'] += 1
                if isinstance(v,dict):
                    counts = counts.setdefault(k, {'c': 0, 'v': {}})
                    counts['c']+=1
                    self.update_Counts(v, counts['v'])
        if isinstance(types,list):
            for k in types:
                if isinstance(k,dict):
                    self.update_Counts(k, counts)
                else:
                    counts = counts.setdefault(k, {'c': 0})
                    counts['c'] += 1



if __name__ == '__main__':
    a = anySimpleType()
    c=XSDTypeDetection()
    import pprint
    pprint.pprint(c._analyse_Column(["5723","5723",'12312312312312']))



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pprint

from io import StringIO
from statistics import mean, median

__author__ = 'jumbrich'
import collections
import functools
import time



class Timer(object):
    GLOBAL_VERBOSE=True
    measures={}

    def __init__(self, verbose=False, key=None, store=True):
        self.verbose = verbose
        self.key=key
        self.store=store

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        end = time.time()
        secs = end - self.start
        msecs = secs * 1000  # millisecs
        if self.verbose and Timer.GLOBAL_VERBOSE:
            print('(%s) elapsed time: %f ms' % (self.key,msecs))
        if self.key and self.store:
            if self.key not in self.__class__.measures:
                self.__class__.measures[self.key]=[]
            if msecs>=0:
                self.__class__.measures[self.key].append(msecs)

    @classmethod
    def printStats(cls):
        print("\n -------------------------")
        print("  Timing stats:" )
        pprint.pprint(cls.getStats())
        print("\n -------------------------")


    @classmethod
    def getStats(cls):
        stats={}
        for m in cls.measures:
            stats[m]={'avg':mean(cls.measures[m]), 'median':median(cls.measures[m]), 'calls':len(cls.measures[m]), 'min':min(cls.measures[m]), 'max':max(cls.measures[m])}
        return stats

    @classmethod
    def to_csv(cls, filename=None):
        import csv
        data = StringIO()
        csvw = csv.writer(data)
        csvw.writerow(['measure, calls', 'min','mean','max','q25','q5','q75'])

        q25, q50, q75=None, None, None
        for m in cls.measures:
            p = cls.measures[m].percentiles
            if cls.measures[m].n >= len(p):
                q25,q50,q75= _sigfigs(p.get(0.25, -1)),_sigfigs(p.get(0.5, -1)), _sigfigs(p.get(0.75, -1))

            csvw.writerow([m, cls.measures[m].n,
                           _sigfigs(cls.measures[m].min), _sigfigs(cls.measures[m].mean), _sigfigs(cls.measures[m].max),
                           q25,q50,q75
                           ])
        if filename:
            with open(filename,'w') as f:
                f.write(data)
        data.seek(0)
        return data.read()


def timer(key=None, verbose=False, *func_or_func_args):
    def wrapped_f(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            with Timer(key=key, verbose=verbose) as t:
                out = f(*args, **kwargs)
            return out
        return wrapped
    if (len(func_or_func_args) == 1
            and isinstance(func_or_func_args[0], collections.Callable)):
        return wrapped_f(func_or_func_args[0])
    else:
        return wrapped_f



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""

A colum/table profiler to detect and extract the potential datetime context of the table
"""


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from dtpattern.dtpattern1 import l2_aggregate, l1_aggregate, translate_all, l3_shared_groups


empty_symbols=[u'NA', u'NAN', u'nan', u'NULL',u'None',u'null']

def is_not_empty(value):
    return not is_empty(value)

def is_empty(value):

    v=value.strip()
    if len(v)==0:
        return True
    return v in empty_symbols

def detect_empty_cell( values ):
    l=[len(v.strip()) for v in values]
    patterns=translate_all(values)
    L2= l2_aggregate(grouped=l1_aggregate(patterns=patterns))
    s_p =l3_shared_groups(L2)
    #print l
    #print L2
    #for k in L2:
    #    print k[0]
    #print s_p


    results=[]
    if min(l)==0:
        results.append('')



if __name__ == '__main__':
    values=[' ',' Wien','Salzburg']
    values = ['-', ' 12', '-12', ' ']

    detect_empty_cell(values)

    print("1234"[4:])



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#http://www.w3.org/TR/2009/WD-xmlschema11-2-20091203/type-hierarchy-200901.longdesc.html
#https://www.w3.org/TR/xmlschema11-2/#built-in-datatypes


class Type(object):


    def detectType(self, value):
        raise NotImplementedError('Must provide implementation in subclass.')

    def type(self):
        return self.__class__.__name__

    def _isType(self, value):
        raise NotImplementedError('Must provide implementation in subclass.')

    def convert(self, value):
        return value

class SingleSubType(Type):

    def __init__(self, subType=None):
        self._subType = subType

    def detectType(self, value):
        results=[]
        if self._isType(value):
            results.append(self.type())
            value= self.convert(value)
            if self._subType is not None:
                res=self._subType.detectType(value)
                if res:
                    #print ("RES",res)
                    if isinstance(res,list):
                        results+=res
                    else:
                        results.append(res)

            return results
        else:
            return None


class MultiSubType(Type):

    def __init__(self, subTypes=None):
        self._subTypes = subTypes

    def detectType(self, value):
        results=[]
        if self._isType(value):
            results.append(self.type())
            value= self.convert(value)
            if self._subTypes is not None:
                for subtype in self._subTypes:
                    res = subtype.detectType(value)
                    if res:
                        if isinstance(res, list):
                            results += res
                        else:
                            results.append(res)

            return results
        else:
            return None








import string

from faker import Factory
fake = Factory.create('it_IT')


## digits

def random_time(size, pattern="%H:%M:%S"):
    return [unicode(fake.time(pattern=pattern))for i in range(0,size)]

def random_iso8601(size):
    return [unicode(fake.iso8601(tzinfo=None))for i in range(0,size)]


def random_date(size, pattern="%Y-%m-%d"):
    return [unicode(fake.date(pattern=pattern))for i in range(0,size)]

def random_number(size, digits=None, fix_len=False):
    return [ unicode(fake.random_number(digits=digits, fix_len=fix_len)) for i in range(0,size)]

def random_lowercase_string(size, length=10):
    return [ unicode(''.join(fake.random_sample(elements=(string.ascii_lowercase), length=length))) for i in range(0,size)]

def random_word(size, length=10):

    return [unicode(''.join(fake.random_sample(elements=(string.ascii_uppercase), length=1)))+unicode(''.join(fake.random_sample(elements=(string.ascii_lowercase), length=length-1))) for i in
            range(0, size)]

def random_isbn13(size, separator='-'):
    return [ unicode(fake.isbn13(separator=separator)) for i in range(0,size) ]
#def random_isbn10(size, separator='-'):
#    return [ unicode(fake.isbn10(separator=separator)) for i in range(0,size)]

def random_isbn10(size, separator='-'):
    return make_pattern(size, fake.isbn10, separator=separator)

def make_pattern(size, f,**fargs):
    return [unicode(f(**fargs)) for i in range(0, size)]



def random_words(size,length=[5,3]):
    res=[]
    for i in range(0,size):
        ws=[]
        for c in length:
            ws.append(unicode(''.join(fake.random_sample(elements=(string.ascii_uppercase), length=1))) + unicode(''.join(fake.random_sample(elements=(string.ascii_lowercase), length=c - 1))))
        res.append(" ".join(ws))
    return res
int_columns=[
    random_number(10,digits=1, fix_len=True),
    random_number(10,digits=2, fix_len=True),
    random_number(10,digits=3), random_word(10, length=10),random_isbn10(10)
]



if __name__ == '__main__':
    print int_columns
    def translate(i):
        return


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from pyyacp.timer import timer, Timer

from pyyacp.profiler import  ColumnProfilerSet, ColumnByCellProfilerSet
from pyyacp.profiler.column_stats_profiler import ColumnStatsProfiler
from pyyacp.profiler.data_type_detection import DataTypeDetection
from pyyacp.profiler.data_type_interpretation import DataTypeInterpretation
from pyyacp.profiler.distributions import CharacterDistributionProfiler, BenfordsLawDistribution
from pyyacp.profiler.fdprofiler import FDProfiler



default_profilers = [ColumnByCellProfilerSet(
        [ColumnPatternProfiler, ColumnStatsProfiler, CharacterDistributionProfiler, BenfordsLawDistribution]),
                 ColumnProfilerSet([DataTypeDetection, DataTypeInterpretation])]

@timer(key="profile_table", verbose=False)
def apply_profilers(table, profilers=default_profilers):
    for i, col in enumerate(table.columns()):
        with Timer(key='column_profilers', verbose=False):
            for profiler in profilers:
                if isinstance(profiler, ColumnProfilerSet):
                    profiler.profile_column(col, table.column_metadata[i])

    with Timer(key='table_profilers', verbose=False):
        for profiler in profilers:
            if not isinstance(profiler, ColumnProfilerSet):
                profiler().profile_table(table)




#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
from io import StringIO

import pandas as pd
import csv

from pyyacp.table_structure_helper import guess_description_lines, _detect_header_lines, _most_common_oneliner, \
    SimpleStructureDetector, AdvanceStructureDetector
from pyyacp.timer import Timer, timer

from pyyacp import YACParserException
import structlog
log = structlog.get_logger()






@timer(key="parse_datatable")
def parseDataTables(yacpParser, url=None, batches=80, max_tables=1, raiseError=True, structure_detector=AdvanceStructureDetector()):
    yacpParser.seek_line(0)
    tables = []
    cur_dt = None
    groups = []

    def grouper(n, iterable):
        it = iter(iterable)
        while True:
            chunk = tuple(itertools.islice(it, n))
            if not chunk:
                return
            yield chunk

    rows = 0
    rows_to_add=[]
    skipped=0
    for g_rows in grouper(batches, yacpParser):
        rows += len(g_rows)

        # analys the shape of the rows
        r_len = list(map(len, g_rows))
        max_len = max(r_len)
        est_colNo = _most_common_oneliner(r_len)
        grouped_L = [ ( k, sum(1 for i in g) ) for k, g in itertools.groupby(r_len)]

        groups += grouped_L

        if len(grouped_L) == 1:
            # perfect, one table in this batch
            if cur_dt is None:
                # we have no table, this is hte first
                comments = structure_detector.guess_description_lines(list(g_rows))
                header = structure_detector.guess_headers(list(g_rows))
                cur_dt = DataTable(yacpParser.meta, est_colNo, comments=comments, headers=header, url=url, id=len(tables))

                pos = len(comments) + len(header)
                rows_to_add.extend(g_rows[pos:])
            elif max_len == cur_dt.no_cols:
                rows_to_add.extend(g_rows)
            else:
                # not the same length, maybe different table , should not happen
                log.warning("NOT IMPLEMENTED", filename=yacpParser.url,
                            msg="not the same length, maybe different table")
        else:
            # lets go over the groups
            # (2,30) -> belongs to old table
            # (0,1)  -> empty line
            # (1,1)  -> comment line -> flag create_new
            # (4,20) -> belongs to new table -> create new table, start parsing at (1,1)

            cur_line = 0
            create_new = False
            for i, group in enumerate(grouped_L):

                if group[0] == 0:   # empty line, skip
                    skipped+=1
                    pass
                elif group[0] == 1 and group[1]<3:
                    # there is a group with one element, that should be the comment lines
                    # also this means a new table
                    if i == len(grouped_L) - 1 and [sum(x) for x in zip(*grouped_L)][1] < batches:
                        # print "SUFFIX COMMENT LINES"
                        log.warning("SUFFIX COMMENT LINES")
                    else:
                        # we have more groups to come, so lets start a new table from this line
                        if not create_new:
                            parse_start = cur_line
                        create_new = True
                else:
                    # a group with more than one column
                    start = None
                    if cur_dt is None or create_new:
                        start = cur_line
                        if create_new:
                            start = parse_start
                    elif group[0] == cur_dt.no_cols:
                        #cur_dt.addRows(g_rows[cur_line:group[1]])
                        rows_to_add.extend(g_rows[cur_line:cur_line+group[1]])
                    else:
                        # seems like a new table
                        if group[1] != 1 or (
                                        i == len(grouped_L) - 1 and
                                        [sum(x) for x in zip(*grouped_L)][1] == batches):
                            # more than one row
                            # OR at the end of the group and still a full batch
                            start = cur_line
                        else:
                            #print ("NOT TREATED", group[0])
                            # if only one row and (at the end of the file or in the middle of a group)
                            pass

                    if start is not None:
                        if cur_dt:
                            with Timer(key="adding {} rows".format(len(rows_to_add)), verbose=True):
                                cur_dt.addRows(rows_to_add)
                                rows_to_add=[]
                            tables.append(cur_dt)

                        _rows=g_rows[start:]
                        comments = structure_detector.guess_description_lines(_rows)
                        header = structure_detector.guess_headers(_rows)

                        cur_dt = DataTable(yacpParser.meta, group[0], comments=comments, headers=header, url=url, id=len(tables))

                        pos = len(comments) + len(header) + start
                        end = cur_line + group[1]

                        rows_to_add.extend(g_rows[pos:end])
                        create_new = False

                cur_line += group[1]

    cur_dt.addRows(rows_to_add)
    rows_to_add = []
    tables.append(cur_dt)

    prev_group = None
    agg_groups = []
    for group in groups:
        if prev_group is not None:
            if prev_group[0] == group[0]:
                # merge
                prev_group = ( prev_group[0], prev_group[1] + group[1])
            else:
                agg_groups.append(prev_group)
                prev_group = group
        else:
            prev_group = group

    agg_groups.append(prev_group)
    log.info("TABLE SHAPE", groups=agg_groups, filename=url)

    if len(tables) > max_tables:
        if raiseError:
            raise YACParserException("Too many tables (#" + str(len(tables)) + ") shapes:" + str(agg_groups))

    log.info("Parsed table", skipped=skipped, tables=len(tables))

    if max_tables==1:
        return tables[0]
    else:
        return tables


class DataTable(object):

    def __init__(self, meta, cols, url, id, comments=None, headers=None):
        self.meta = meta
        self.no_cols = cols
        self.no_rows = 0
        self.uri=url
        self.id=id

        self.comments = comments if comments else []
        self.header_rows = headers if headers else []
        self.id = None
        self.columnIDs = []
        if len(self.header_rows)==1:
            for i, v in enumerate(self.header_rows[0]):
                if len(v)>0:
                    self.columnIDs.append(v)
                else:
                    self.columnIDs.append('empty{}'.format(i))
        elif len(self.header_rows) > 1:

            for i in range(0, self.no_cols):
                added = False
                for h_row in self.header_rows:
                    if len(h_row[i])>0:
                        self.columnIDs.append(h_row[i])
                        added=True
                        break
                if not added:
                    self.columnIDs.append('empty{}'.format(i+1))
        else:
            self.columnIDs =['col{}'.format(i) for i in range(1, self.no_cols+1)]

        self.column_metadata={i:{}for i in range(0,len(self.columnIDs))}

        ### make sure that we do not have duplicate column names
        newlist = []
        for i, v in enumerate(self.columnIDs):
            totalcount = self.columnIDs.count(v)
            count = self.columnIDs[:i].count(v)
            newlist.append(v + str(count + 1) if totalcount > 1 else v)
        self.columnIDs=newlist
        self.data=pd.DataFrame(columns=self.columnIDs)

    def table_structure(self):
        b={
            'uri': self.uri,
            'rows':self.no_rows,
            'columns':self.no_cols,
            'headers':self.header_rows,
            'comments': self.comments,
            'comment_lines': len(self.comments),
            'header_lines': len(self.header_rows)
        }
        return b

    def table_profile(self):

        meta = {k:v for k, v in self.meta.items()}
        if 'table_tane' in self.meta:
            meta['fd']=self.meta['table_tane']
        return meta

    def addRows(self, rows):
        try:
            df = pd.DataFrame(list(rows), columns=self.columnIDs)
            self.data=pd.concat([self.data,df])
            #print ('addRows {} {}'.format(len(rows), self.data.shape))
            self.no_rows=self.data.shape[0]
        except Exception as e:
            print(e)

    def rows(self):
        return [row for row in self.rowIter()]

    def rowIter(self):
        for row in self.data.itertuples():
            yield list(row[1:])


    def columnIter(self):
        for colName in self.columnIDs:
            yield self.data[colName].tolist()

    def columns(self):
        return [collist for collist in self.columnIter()]

    #def columns(self):
    #    return [ {'data':col,'meta':self.column_metadata[i]} for i, col in enumerate(self.columnIter())  ]



    def colum_profiles(self):

        d={}
        for i in range(0, self.no_cols):
            dd=d.setdefault('col{}'.format(i+1),{})
            for a,h in enumerate(self.header_rows):
                dd['header{}'.format(a)]=h[i]

            for k,v in self.column_metadata[i].items():
                dd[k]=v

        return pd.DataFrame(d)


    def generate(self, delimiter=',', newline='\n', header=True, comments=True):
        """
        :param delimiter: The delimiter symbol. The default is ",".
        :param newline: The line separator. The default is "\n".
        :param header: There will be a header in the output (if no header detected then col0,col1, ...)
        :param commentPrefix: The prefix for comments in the first rows. If false, any comments will be ignored.
        :return: A string representation of the CSV table
        """
        csvstream = StringIO()
        w = csv.writer(csvstream,  delimiter=str(delimiter), lineterminator=str(newline))
        c=0
        # write description lines at top
        if comments:
            for line in self.comments:
                w.writerow(line)
                c+=1
        if header:
            for header in self.header_rows:
                w.writerow([str(h) for h in header])
                c += 1
        for row in self.rowIter():
            w.writerow([str(r) for r in row])
            c += 1
        return csvstream.getvalue()


    def print_summary(self, column_raw=True):
        print ('_' * 30, 'TABLE META', '_' * 30)
        for k,v in self.table_profile().items():
            print (k,v)
        print ('_' * 30, 'TABLE STRUCTURE', '_' * 30)
        for k,v in self.table_structure().items():
            print (k,v)
        print ('_' * 30, 'COLUMN PROFILE', '_' * 30)
        print (self.colum_profiles())

        print('_' * 30, 'DATA {}'.format(self.data.shape), '_' * 30)
        print(self.data.head(5))

        #if column_raw:
        #    for i,c in enumerate(self.columns()):
        #        print ('COL',i,set(c))
        #    print('*' * 80)






#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import  pyyacp.datatable as datatable
from pyyacp.table_structure_helper import AdvanceStructureDetector

SAMPLES_PATH = "sample_csvs"
from os import listdir
from os.path import isfile, join
onlyfiles = [join(SAMPLES_PATH, f) for f in listdir(SAMPLES_PATH) if isfile(join(SAMPLES_PATH, f))]



for csv_file in onlyfiles:
    if 'multi_head_milti_table.csv' not in csv_file:
        continue

    from pyyacp.pyyacp import YACParser

    yacp = YACParser(filename=csv_file,structure_detector = AdvanceStructureDetector(),sample_size=1800)
    print yacp


    tables=datatable.parseDataTables(yacp, url='http://example.org/test', max_tables=10)

    for table in tables:
        print table.data.shape
        print table.data.head(5)


        print 'Comments', table.comments
        print 'Headers', table.header_rows



# -*- coding: utf-8 -*-
import re, string, timeit


input= "test_123-TEST"

digit_table = string.maketrans(string.digits,"1"*len(string.digits))
ascii_lower_table = string.maketrans(string.ascii_lowercase,"a"*len(string.ascii_lowercase))
ascii_upper_table = string.maketrans(string.ascii_uppercase,"A"*len(string.ascii_uppercase))

in_out={
    string.digits:'1',
    string.ascii_lowercase:'a',
    string.ascii_uppercase:'A'
}
in_out2={
    string.digits:'1',
    string.ascii_lowercase:'a',
    string.ascii_uppercase:'A',
    'ä':'a'
}


def make_trans_table(in_out):
    input=''
    output=''
    for i, o in in_out.items():
        input+=i
        output+=o*len(i)
    return string.maketrans(input, output)

def make_trans_table_unicode(in_out):
    translate_table={}
    for i, o in in_out.items():
        for char in i.decode('utf-8'):
            translate_table[ord(char)]=unicode(o)

    return translate_table

all_table = make_trans_table(in_out)
all_table_unicode=make_trans_table_unicode(in_out2)

def translate_digits(s):
    return s.translate(digit_table)
def translate_ascii_lower(s):
    return s.translate(ascii_lower_table)
def translate_ascii_upper(s):
    return s.translate(ascii_upper_table)

def translate(s):
    s=translate_digits(s)
    s=translate_ascii_lower(s)
    s=translate_ascii_upper(s)
    return s

def translate1(s):
    return s.translate(all_table)

def translate5(s):
    return s.translate(all_table_unicode)


def translate2(s):
    s=re.sub(r'\d', '1', s)
    s=re.sub(r'[A-Z]', 'A', s)
    s=re.sub(r'[a-z]', 'a', s)
    return s


num_dict = {'123456': "1",
                'test': "a",
                'TEST': "A",
                '_':'_',
                '-': '-'}
def translate4(s):

    string = ""
    for c in s:
        for k in num_dict.keys():
            if c in k:
                string += num_dict[k]
    return string



if __name__ == '__main__':

    print "input",input
    print "translate", translate(input)
    print "translate1", translate1(input)
    print "translate2", translate2(input)
    print "translate4", translate4(input)
    print "translate5", translate5('ä'.decode('utf-8'))



    #print "sets      :", timeit.Timer('f(s)', 'from __main__ import s,test_set as f').timeit(1000000)
    print "translate :", timeit.Timer('f(input)', 'from __main__ import input,translate as f').timeit(1000000)
    print "translate1   :", timeit.Timer('f(input)', 'from __main__ import input,translate1 as f').timeit(1000000)
    print "translate2   :", timeit.Timer('f(input)', 'from __main__ import input,translate2 as f').timeit(1000000)
    print "translate4   :", timeit.Timer('f(input)', 'from __main__ import input,translate4 as f').timeit(1000000)


#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import traceback

import datetime
import requests

from collections import defaultdict
import time
import structlog
log =structlog.get_logger()

odpwAPI="http://data.wu.ac.at/portalwatch/api/v1/"
datamonitorAPI="http://csvengine.ai.wu.ac.at/datamonitor/api/v1/"

base='sample_csvs'

def getCurrentSnapshot():
    now = datetime.datetime.now()
    y=now.isocalendar()[0]
    w=now.isocalendar()[1]
    sn=str(y)[2:]+'{:02}'.format(w)

    return sn

def getURLperPortal(odpwAPI,portal, snapshot, format):
    return odpwAPI+"portal/{}/{}/resources?format={}".format(portal, snapshot, format)

def getURLInfo(datamonitorAPI,url):
    return datamonitorAPI+"info/{}".format(url)

def url_iter(portalID, snapshot=getCurrentSnapshot()):
    requestURL = getURLperPortal(odpwAPI, portalID, snapshot, 'csv')
    log.info("ODPW", url=requestURL)
    res = requests.get(requestURL)
    log.info("ODPW", url=requestURL, status=res.status_code)

    if res.status_code == 200:
        data=res.json()
        log.info("ODPW SUCCESS", portal=portalID,snapshot=snapshot, format=format, urls=len(data))
        for url in data:
            yield url

def csvContent_iter(portalID, snapshot=None, useDatamonitor=False):
    if snapshot==None:
        snapshot=getCurrentSnapshot()
    urls=[url for url in url_iter(portalID, snapshot)]
    print (len(urls))
    for url in urls:
        print (url)
        content=None
        if useDatamonitor:
            inforequestURL = getURLInfo(datamonitorAPI, url)
            info = requests.get(inforequestURL)
            log.info("DATAMONITOR SUCCESS", url=inforequestURL, status=info.status_code)
            if info.status_code == 200:
                data = info.json()
                print (data)
                if 'status' not in data or data['status'] == '404':
                    indexFromURL = True
            content=None

        if content:
            filename=cached(url, content)
        else:
            filename = cached(url)

        yield url, filename




def cached(uri, content=None):
    import hashlib
    m = hashlib.md5()
    m.update(uri.encode('utf-8'))
    hashid= m.hexdigest()
    filename=os.path.join(base,hashid+".csv")
    if not os.path.isfile(filename):
        if not content:
            resp = requests.get(uri)
            if resp.status_code == 200:
                content=resp.content
        if content:
            with open(filename, "wb") as handle:
                handle.write(content)
        else:
            return None

    return filename




from pyyacp import datatable
from pyyacp.pyyacp import YACParser
from pyyacp.profiler.fdprofiler import TANEAlgorithm, FDProfiler
from pyyacp.table_structure_helper import AdvanceStructureDetector
from pyyacp.testing.csv_iterator import csvContent_iter
from pyyacp.timer import Timer

if __name__ == '__main__':
    tane=TANEAlgorithm()


    cols={'A':[i for i in range(0,100)],
          'B':[i%2 for i in range(0,100)],
          'C':[i%3 for i in range(0,100)],
          'D': [i % 4 for i in range(0, 100)],
          'E': [i % 2 for i in range(0, 100)]
          }
    with Timer(verbose=True):
        print tane.analyse_cols(cols.values())

    portalID = 'data_wu_ac_at'

    profilers = [FDProfiler()]  # ,ColumnRegexProfiler()]#,XSDTypeDetection()]
    c = 0
    for uri, csv_file in csvContent_iter(portalID):
        c += 1
        print "{}, {} -> {}".format(c, uri, csv_file)
        try:
            from pyyacp.yacp import YACParser

            yacp = YACParser(filename=csv_file, sample_size=1800, structure_detector=AdvanceStructureDetector(),
                             verbose=True)

            table = datatable.parseDataTables(yacp, url=uri)
            print table.data.head(5)

            profiler_engine.profile(table, profilers=profilers)

            print 'Comments', table.comments
            print 'Headers', table.header_rows

            for i in range(0, table.no_cols):
                print "Column", i, table.header_rows[0][i]
                for p in profilers:
                    if p.key.startswith("col"):
                        if len(table.meta[p.key]) >= i:
                            print ' [{}] {} '.format(p.id, table.meta[p.key][i])
            for p in profilers:
                if p.key.startswith("table"):
                    print '[{}] {} '.format(p.id, table.meta[p.key])

            print table.meta



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import math
import numpy as np

from collections import defaultdict

from pyyacp.profiler import ColumnProfiler, ColumnByCellProfiler
from pyyacp.profiler import is_not_empty
from pyyacp.profiler.data_type_detection import FLOAT, UNICODE
from pyyacp.profiler.data_type_detection import INT
from pyyacp.timer import timer


class CharacterDistributionProfiler(ColumnProfiler,ColumnByCellProfiler):

    def __init__(self):
        super(CharacterDistributionProfiler, self).__init__('cdp', 'c_dist')
        self.dv = defaultdict(int)

    @timer(key='cdp_column')
    def profile_column(self, column, meta):
        return self._profile_column(column, meta)

    def result(self):
        return self._compile_stats()

    def accept(self, cell):
        if is_not_empty(cell):
            for c in cell:
                self.dv[c] += 1

        #self.values.append(cell)

    def _profile_column(self, values, meta):
        data_type = meta['data_type']
        if data_type == UNICODE:
            for v in filter(is_not_empty,values):
                for c in v:
                    self.dv[c] += 1
        return self._compile_stats()

    def _compile_stats(self):
        if len(self.dv)>0:
            return {'dist':dict(self.dv)}
        else:
            return {'dist': None}

def benford_law():
    return [math.log10(1 + 1 / float(i)) * 100.0 for i in range(1, 10)]

def get_leading_number(cell):
    numbers = "123456789"
    l = list(map(int, filter(lambda x: x in numbers, cell)))
    return l[0] if len(l)>0 else 0

def chisq_stat(O, E):
    return sum([(o - e) ** 2 / e for (o, e) in zip(O, E)])

def passChi(chi):
	# degrees of freedom always 8 and assuming 95% confidence
	# from: https://en.wikipedia.org/wiki/Chi-squared_distribution
	# 95% conf: 15.51

	# Null hypothesis H_0: The first digits in the index minute price changes follow the Benford's law
	# Hypothesis H_1: The first digits in the index minute price changes do not follow the Benford's law}
	if (chi<15.51):
		return True
	else:
		return False

def getExpectedBenfordCount(size):
	return list(map(lambda x:x*0.01*size, benford_law()))

class BenfordsLawDistribution(ColumnProfiler, ColumnByCellProfiler):

    def __init__(self):
        super(BenfordsLawDistribution, self).__init__('cdb', 'benford')
        self.dv = {i:0 for i in range(1,10)}

    @timer(key='cdb_column')
    def profile_column(self, column, meta):
        return self._profile_column(column, meta)

    def result(self):
        return self._compile_stats()

    def accept(self, cell):
        digit=get_leading_number(cell)
        if digit != 0:
            self.dv[digit] +=1

    def _profile_column(self, values, meta):
        data_type = meta['data_type']
        if data_type == INT or data_type == FLOAT:
            x = map(get_leading_number, values)
            for k in range(1, 10):
                self.dv[k] = x.count(k)
        return self._compile_stats()

    def _compile_stats(self):
        if sum(self.dv.values())>0:
            d = np.array([ self.dv.get(k,0) for k in range(1, 10)])
            y = getExpectedBenfordCount(sum(d))

            chi= chisq_stat(d, y)

            dd={}
            total=sum(d)*1.0
            for k,v in self.dv.items():
                dd[k]=v/(total)

            return {'dist': dd, 'is':passChi(chi), 'chi':chi}
        else:
            return {'dist': None, 'is': False, 'chi': None}

#
#
#
# ### BENCHMARKING
#
# c = None
# def dist_counter(values):
#     global c
#     c=Counter("".join(values))
#     return c
#
# r = defaultdict(int)
# def dist_default(values):
#     global r
#     r = defaultdict(int)
#     for v in values:
#         for c in v:
#             r[c]+=1
#     return r
#
#
# from faker import Factory
# fake = Factory.create()
# inputs=[unicode(fake.name()) for i in range(0,1000)]
# print inputs
# print dist_default(inputs)
# print dist_counter(inputs)
# if __name__ == '__main__':
#
#
#     #print s
#
#     # print "sets      :", timeit.Timer('f(s)', 'from __main__ import s,test_set as f').timeit(1000000)
#     print "translate :", timeit.Timer('f(inputs)', 'from __main__ import inputs,dist_counter as f').timeit(1000)
#     print "translate :", timeit.Timer('f(inputs)', 'from __main__ import inputs,dist_default as f').timeit(1000)
#     #print "translate1   :", timeit.Timer('f(input)', 'from __main__ import input,translate1 as f').timeit(1000000)
#     #print "translate2   :", timeit.Timer('f(input)', 'from __main__ import input,translate2 as f').timeit(1000000)
#     #print "translate4   :", timei t.Timer('f(input)', 'from __main__ import input,translate4 as f').timeit(1000000)
#     print c
#     print r